#######################################################
# 
# FederationClientService.py
# Python implementation of the Class FederationClientService
# Generated by Enterprise Architect
# Created on:      29-Dec-2020 8:10:38 AM
# Original author: natha
# 
#######################################################
from FreeTAKServer.controllers.configuration.MainConfig import MainConfig
from FreeTAKServer.controllers.configuration.types import Types
from FreeTAKServer.controllers.services.federation.handlers import StopHandler, DisconnectHandler, ConnectHandler, SendDataHandler, SendConnectionDataHandler, SendDisconnectionDataHandler
from FreeTAKServer.model.protobufModel.fig_pb2 import FederatedEvent
from FreeTAKServer.controllers.services.service_abstracts import ServerServiceInterface, ServiceBase
from multiprocessing import Pipe as multiprocessingPipe
from FreeTAKServer.model.federate import Federate
import selectors
import socket
from typing import Tuple
import ssl
import codecs
from FreeTAKServer.controllers.serializers.protobuf_serializer import ProtobufSerializer
from FreeTAKServer.controllers.serializers.xml_serializer import XmlSerializer
from FreeTAKServer.controllers.XMLCoTController import XMLCoTController
from FreeTAKServer.model.SpecificCoT.SendOther import SendOther
from FreeTAKServer.model.FTSModel.Event import Event
from FreeTAKServer.model.ClientInformation import ClientInformation
from FreeTAKServer.model.SpecificCoT.SendDisconnect import SendDisconnect
from FreeTAKServer.controllers.DatabaseControllers.DatabaseController import DatabaseController
from FreeTAKServer.controllers.CreateLoggerController import CreateLoggerController
logger = CreateLoggerController("FederationClient").getLogger()

class FederationClientServiceController(ServerServiceInterface, ServiceBase):
#class FederationClientServiceController:
    """A service which controllers the connection too and transfer of data with
    federated servers.
    """

    def __init__(self):
        self._define_responsibility_chain()
        self.pipe = None
        self.federates: {str: Federate} = {}
        self.sel = selectors.DefaultSelector()

    def _send_connected_clients(self, connection):
        clients = self.db.query_user()
        for client in clients:
            try:
                proto_obj = FederatedEvent()
                proto_obj.contact.uid = str(client.uid)
                proto_obj.contact.callsign = str(client.CoT.detail.contact.callsign)
                proto_obj.contact.operation = 1
                proto_str = proto_obj.SerializeToString()
                header = self._generate_header(len(proto_str))
                connection.send(header + proto_str)
            except Exception as e:
                logger.warning("error thrown sending federate data to newly connected federate " + str(e))
                continue

    def _create_context(self):
        self.context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        self.context.load_cert_chain(MainConfig.federationCert, MainConfig.federationKey,
                                     password=MainConfig.federationKeyPassword)
        self.context.set_ciphers('DEFAULT@SECLEVEL=1')

    def _define_responsibility_chain(self):
        self.m_StopHandler = StopHandler()

        self.m_ConnectHandler = ConnectHandler()
        self.m_ConnectHandler.setNextHandler(self.m_StopHandler)

        self.m_DisconnectHandler = DisconnectHandler()
        self.m_DisconnectHandler.setNextHandler(self.m_ConnectHandler)

        self.m_SendDataHandler = SendDataHandler()
        self.m_SendDataHandler.setNextHandler(self.m_DisconnectHandler)

        self.m_SendDisconnectionHandler = SendDisconnectionDataHandler()
        self.m_SendDisconnectionHandler.setNextHandler(self.m_SendDataHandler)

        # first handler in chain of responsibility and should be called first
        self.m_SendConnectionHandler = SendConnectionDataHandler()
        self.m_SendConnectionHandler.setNextHandler(self.m_SendDisconnectionHandler)

    def main(self):
        from lxml import etree
        import time
        while True:
            time.sleep(0.1)
            command = self.receive_command_data(self.pipe)
            if command:
                try:
                    self.m_SendConnectionHandler.Handle(self, command)
                except Exception as e:
                    logger.debug("exception in command chain "+str(e))
            else:
                pass

            data = self.receive_data_from_federate(1)
            if data:
                for protobuf_object in data:
                    # TODO: clean all of this up as it's just a PoC

                    try:
                        detail = etree.fromstring(protobuf_object.event.other)
                        protobuf_object.event.other = ''
                        fts_obj = ProtobufSerializer().from_format_to_fts_object(protobuf_object, Event.FederatedCoT())
                        specific_obj = SendOther()
                        event = XmlSerializer().from_fts_object_to_format(fts_obj)
                        xmlstring = event
                        xmlstring.find('detail').remove(xmlstring.find('detail').find('remarks'))
                        xmlstring.find('detail').extend([child for child in detail])
                        # specific_obj.xmlString = etree.tostring(xmlstring)
                        print(etree.tostring(xmlstring))
                        specific_obj.xmlString = etree.tostring(xmlstring)
                        self.pipe.send(specific_obj)
                    except Exception as e:
                        pass
                    """if isinstance(SpecificCoTObj, SendOtherController):
                        detail = protobuf_object.event.other
                        protobuf_object.event.other = ''
                        fts_obj = ProtobufSerializer().from_format_to_fts_object(protobuf_object, Event.Other())
                        protobuf_object.event.other = detail
                        SpecificCoTObj.object = fts_obj
                        SpecificCoTObj.Object =
                    else:
                        fts_obj = ProtobufSerializer().from_format_to_fts_object(protobuf_object, SpecificCoTObj().object)
                        self.pipe.send(data)"""
            else:
                pass

    def connect_to_server(self, server_vars: Tuple[str, str]) -> None:
        try:
            federate_db_obj = self.db.query_Federation(f'id == "{server_vars[0]}"')[0]
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
            ssock = self.context.wrap_socket(sock, server_hostname=federate_db_obj.address)
            ssock.settimeout(10)
            ssock.connect((str(federate_db_obj.address), int(federate_db_obj.port)))
            ssock.setblocking(False)
            federate = Federate()
            federate.uid = server_vars[0]
            federate.addr = federate_db_obj.address
            federate.conn = ssock
            federate.name = federate_db_obj.name
            events = selectors.EVENT_READ | selectors.EVENT_WRITE
            self.sel.register(ssock, events, federate)
            self.federates[server_vars[0]] = federate
            self._send_connected_clients(ssock)
            self.db.create_ActiveFederation(id = federate_db_obj.id, address = federate_db_obj.address,
                                            port = federate_db_obj.port, initiator = "Self")
            self.db.update_Federation({"lastError": None}, query=f'id == "{federate_db_obj.id}"')
            return None
        except Exception as e:
            try:
                self.db.remove_ActiveFederation(f'id == "{server_vars[0]}"')
            except Exception as e:
                logger.warning("exception thrown removing outgoing federation from DB "+str(e))
            logger.warning("exception thrown creating new federation "+str(e))
            try:
                self.db.update_Federation({"status": "Disabled", "lastError": str(e)}, query=f'id == "{server_vars[0]}"')
            except Exception as e:
                logger.warning("exception thrown updating federate in db "+str(e))
    def disconnect_client(self, id: str) -> None:
        try:
            try:
                federate = self.federates[id]
            except Exception as e:
                logger.warning("federate array has no item with uid "+str(id)+" federates array is len " + str(len(self.federates)))
                return None
            try:
                federate.conn.close()
                self.sel.unregister(federate.conn)
                del(self.federates[federate.uid])
            except Exception as e:
                logger.warning("exception thrown disconnecting client "+str(e))
            try:
                self.db.remove_ActiveFederation(f'id == "{federate.uid}"')
            except Exception as e:
                logger.warning("exception thrown removing outgoing federation from DB "+str(e))
            return None
        except Exception as e:
            logger.warning("exception thrown accessing client for disconnecting client " + str(e))

    def receive_data_from_federate(self, timeout):
        """called whenever data is available from any federate and immediately proceeds to
        send data through process pipe
        """
        dataarray = []
        if self.federates:
            events = self.sel.select(timeout)
            for key, mask in events:
                conn = key.fileobj
                try:
                    header = conn.recv(4)
                except Exception as e:
                    continue
                if header:
                    try:
                        buffer = self._get_header_length(header)
                        raw_protobuf_message = conn.recv(buffer)
                        print(raw_protobuf_message)
                        protobuf_object = FederatedEvent()
                        protobuf_object.ParseFromString(raw_protobuf_message)
                        dataarray.append(protobuf_object)
                    except Exception as e:
                        conn.recv(10000)
                        continue
                else:
                    self.disconnect_client(key.data.uid)
            return dataarray
        else:
            return None

    def _get_header_length(self, header):
        return int.from_bytes(header, 'big')

    def _generate_header(self, contentlength):
        return contentlength.to_bytes(4, byteorder="big")

    def send_data_to_clients(self, data):
        from lxml import etree
        try:
            if self.federates:
                xmlstring = data.xmlString
                detail = etree.fromstring(xmlstring).find('detail')
                if detail:
                    protobuf = ProtobufSerializer().from_fts_object_to_format(data.modelObject)
                    try:
                        protobuf.event.other = etree.tostring(detail)
                        protobufstring = protobuf.SerializeToString()
                        header = self._generate_header(len(protobufstring))
                        protobufstring = header + protobufstring
                        print(protobufstring)
                    except Exception as e:
                        logger.warning("creating protobuf message failed "+str(e))
                        return None
                    for client in self.federates.values():
                        client.conn.send(protobufstring)
                else:
                    return None
            else:
                return None
        except Exception as e:
            import traceback
            trace = traceback.format_exc()
            logger.warning("sending data to federates failed "+str(e))

    def send_connection_data(self, CoT: ClientInformation) -> None:
        try:
            if self.federates:
                logger.debug("connection data received in send_connection_data")
                proto_obj = FederatedEvent()
                proto_obj.contact.uid = str(CoT.modelObject.uid)
                proto_obj.contact.callsign = str(CoT.modelObject.detail.contact.callsign)
                proto_obj.contact.operation = 1
                proto_str = proto_obj.SerializeToString()
                header = self._generate_header(len(proto_str))
                for fed in self.federates.values():
                    fed.conn.send(header + proto_str)
                return None

            else:
                return None

        except Exception as e:
            logger.warning("exception throw sending new connection data to federates " + str(e))
            return None

    def send_disconnection_data(self, CoT: SendDisconnect):
        if self.federates:
            proto_obj = FederatedEvent()
            proto_obj.contact.uid = str(CoT.modelObject.detail.link.uid)
            proto_obj.contact.callsign = str(CoT.modelObject.detail.link.type)
            proto_obj.contact.operation = 4
            proto_str = proto_obj.SerializeToString()
            header = self._generate_header(len(proto_str))
            for fed in self.federates.values():
                fed.conn.send(header + proto_str)
        else:
            return None

    def start(self, pipe):
        self.db = DatabaseController()
        self.pipe = pipe
        self._create_context()
        print('started federation federate service')
        self.main()

    def stop(self):
        pass

if __name__ == "__main__":
    #FederationClientServiceController()._get_header_length(b'\x00\x00\x03>')
    """from FreeTAKServer.controllers.SpecificCoTControllers.SendOtherController import SendOtherController
    from FreeTAKServer.model.RawCoT import RawCoT
    a = FederationClientService()._get_header_length(b'\x00\x00\x02\x05')
    pipe1, pipe2 = multiprocessingPipe(True)
    pipe2.send(('3.96.240.142', 9000, 'testconn'))
    cot = RawCoT()
    cot.xmlString = '<event version="2.0" uid="ANDROID-R5CN70EYKQH" type="a-f-G-U-C" how="h-e" start="2020-12-24T18:16:22.325Z" time="2020-12-24T18:16:22.325Z" stale="2020-12-24T18:22:37.325Z"><detail><__group name="Teal" role="Team Member"/><status battery="76"/><takv version="4.2.0.4 (47e136dd).1607456856-CIV" platform="ATAK-CIV" device="SAMSUNG SM-N986U" os="29"/><track course="159.1462509079387" speed="0.0"/><contact callsign="SPAC3SLOTH" endpoint="*:-1:stcp" /><uid Droid="SPAC3SLOTH"/><precisionlocation altsrc="GPS" geopointsrc="GPS"/></detail><point le="9999999.0" ce="11.0" hae="178.84407323983876" lon="-76.675505" lat="39.664392"/></event>'
    obj = SendOtherController(RawCoT=cot)
    pipe2.send(obj.getObject())
    FederationClientService().start(pipe1)
    while True:
        data = pipe2.recv()
        print(data.xmlString)"""



class AbstractPythonSerializer:
    def __init__(self, object):
        self.pythonobject = object

    def serializeObject(self):
        for item in vars(self.pythonobject):

            if isinstance(getattr(self.pythonobject, item), list):
                self.list(item)
            elif isinstance(getattr(self.pythonobject, item), dict):
                self.dict(item)
            elif (str(type(getattr(self.pythonobject, item))).split("'")[1]).split('.')[0] == 'FreeTAKServer':
                self.object(item)
            else:
                self.other(item)

    def list(self, obj):
        pass

    def dict(self, obj):
        pass

    def object(self, obj):
        pass

    def other(self, obj):
        pass

if __name__ == "__main__":
    from FreeTAKServer.model.ExCheck.Checklists.Checklist import Checklist
    AbstractPythonSerializer(Checklist()).serializeObject()


#######################################################
# 
# Event.py
# Python implementation of the Class Event
# Generated by Enterprise Architect
# Created on:      26-Sep-2020 12:18:51 PM
# Original author: natha
# 
#######################################################
from sqlalchemy import Column
from FreeTAKServer.model.SQLAlchemy.Root import Base, Root
from FreeTAKServer.model.SQLAlchemy.CoTTables.Detail import Detail
from FreeTAKServer.model.SQLAlchemy.CoTTables.Point import Point
from sqlalchemy import String
from sqlalchemy.orm import relationship


class Event(Base, Root):
# default constructor  def __init__(self):  

    __tablename__ = "Event"
    uid = Column(String, primary_key=True, unique=True)
    how = Column(String)
    start = Column(String)
    type = Column(String)
    version = Column(String)
    point = relationship("Point", uselist=False, backref="Event", cascade="all, delete, merge, delete-orphan")
    detail = relationship("Detail", uselist=False, backref="Event", cascade="all, delete, merge, delete-orphan")


#######################################################
# 
# ReceiveConnectionsProcessController.py
# Python implementation of the Class ReceiveConnectionsProcessController
# Generated by Enterprise Architect
# Created on:      21-May-2020 9:35:07 AM
# Original author: Natha Paquette
# 
#######################################################
from FreeTAKServer.model.ReceiveConnectionsProcess import ReceiveConnectionsProcess

class ReceiveConnectionsProcessController:
    def __init__(self):  
        self.ReceiveConnectionsProcess = ReceiveConnectionsProcess()

    def InstantiateModel(self, ReceiveConnectionProcess):
        self.ReceiveConnectionsProcess.receiveConnections = ReceiveConnectionProcess
        return self.ReceiveConnectionsProcess.receiveConnections



from FreeTAKServer.controllers.configuration.MainConfig import MainConfig

class CoTServiceVariables:
    def __init__(self):
        self.CoTServiceIP = "0.0.0.0"
        self.CoTServicePort = MainConfig.CoTServicePort
        self.CoTServiceStatus = ""


#######################################################
# 
# UserConfiguration.py
# Python implementation of the Class UserConfiguration
# Generated by Enterprise Architect
# Created on:      21-Sep-2020 10:28:11 PM
# Original author: natha
# 
#######################################################
from PSM.Column import Column
from PSM.Root import Root

class UserConfiguration(Root):
# default constructor  def __init__(self):  

    Group = ""    
    Password = ""    
    UserName = ""    


from sqlalchemy import Column, ForeignKey
from FreeTAKServer.model.SQLAlchemy.Root import Base, Root
from sqlalchemy import String
from sqlalchemy import Integer
from FreeTAKServer.model.SQLAlchemy.ExCheckKeywords import ExCheckKeywords
from sqlalchemy.orm import relationship

class ExCheckData(Base):
    __tablename__ = 'ExCheckData'
    PrimaryKey = Column(ForeignKey("ExCheck.PrimaryKey"), primary_key = True, autoincrement=True)
    filename = Column(String)
    mimeType = Column(String)
    name = Column(String)
    submissionTime = Column(String)
    submitter = Column(String)
    uid = Column(String)
    hash = Column(String, unique=True)
    size = Column(Integer)
    tool = Column(String)
    keywords = relationship("ExCheckKeywords", uselist=False, backref="ExCheckData", cascade="all, delete")


#######################################################
# 
# Status.py
# Python implementation of the Class Status
# Generated by Enterprise Architect
# Created on:      26-Sep-2020 9:41:29 PM
# Original author: natha
# 
from sqlalchemy import Column
from FreeTAKServer.model.SQLAlchemy.Root import Base
from sqlalchemy import String
from sqlalchemy import ForeignKey


class Status(Base):
# default constructor  def __init__(self):  
    __tablename__ = "Status"
    PrimaryKey = Column(ForeignKey("Detail.PrimaryKey"), primary_key=True)
    # % of the battery on the phone
    battery = Column(String)
    # probably boolean to determine if ready or not
    readiness = Column(String)


import xml.etree.ElementTree as etree


class BasicModelInstantiate:
    def __init__(self, root, modelObject):
        try:
            self.modelObject = modelObject
            root = root.encode()
            self.event = etree.XML(root)
            self.establishVariables()
            self.eventAtrib()
            self.pointAtrib()
            self.takvAtrib()
            self.contactAtrib()
            self.uidAtrib()
            self.precisionlocationAtrib()
            self.groupAtrib()
            self.statusAtrib()
            self.trackAtrib()
        except Exception as e:
            print(e)

    def establishVariables(self):
        self.detail = self.event.find('detail')
        self.point = self.event.find('point')
        self.takv = self.detail.find('takv')
        self.contact = self.detail.find('contact')
        self.uid = self.detail.find('uid')
        self.precisionlocation = self.detail.find('precisionlocation')
        self.group = self.detail.find('__group')
        self.status = self.detail.find('status')
        self.track = self.detail.find('track')
        self.link = self.detail.find('link')
        self.emergency = self.detail.find('emergency')

    def eventAtrib(self):
        self.modelObject.uid = self.event.attrib['uid']
        self.modelObject.type = self.event.attrib['type']
        self.modelObject.how = self.event.attrib['how']

    def pointAtrib(self):
        self.modelObject.Point.lat = self.point.attrib['lat']
        self.modelObject.Point.lat = self.point.attrib['lon']
        self.modelObject.Point.lat = self.point.attrib['hae']
        self.modelObject.Point.lat = self.point.attrib['ce']
        self.modelObject.Point.lat = self.point.attrib['le']

    def takvAtrib(self):
        self.modelObject.detail.Takv.os = self.takv.attrib['os']
        self.modelObject.detail.Takv.device = self.takv.attrib['device']
        self.modelObject.detail.Takv.platform = self.takv.attrib['platform']
        self.modelObject.detail.Takv.version = self.takv.attrib['version']

    def contactAtrib(self):
        try:
            self.modelObject.detail.Contact.endpoint = self.contact.attrib['endpoint']
        except:
            pass
        try:
            self.modelObject.detail.Contact.callsign = self.contact.attrib['callsign']
        except:
            pass
    def uidAtrib(self):
        self.modelObject.detail.Uid.Droid = self.uid.attrib['Droid']
    
    def precisionlocationAtrib(self):
        try:
            self.modelObject.detail.precisionlocation.altsrc = self.precisionlocation.attrib['altsrc']
        except:
            pass
        try:
            self.modelObject.detail.precisionlocation.geopointsrc = self.precisionlocation.attrib['geopointsrc']
        except:
            pass
    def groupAtrib(self):
        self.modelObject.detail.Group.role = self.group.attrib['role']
        self.modelObject.detail.Group.name = self.group.attrib['name']

    def statusAtrib(self):
        self.modelObject.detail.Status.battery = self.status.attrib['battery']
    
    def trackAtrib(self):
        self.modelObject.detail.Track.course = self.track.attrib['course']
        self.modelObject.detail.Track.speed = self.track.attrib['speed']
    
    def linkAtrib(self):
        try:
            self.modelObject.detail.link.uid = self.link.attrib['uid']
        except:
            pass
        try:
            self.modelObject.detail.link.type = self.link.attrib['type']
        except:
            pass
        try:
            self.modelObject.detail.link.relation = self.link.attrib['relation']
        except:
            pass
    def emergencyAtrib(self):
        try:
            self.modelObject.detail.emergency.type = self.emergency.attrib['type']
        except:
            pass
        
        try:
            self.modelObject.detail.emergency.cancel = self.emergency.attrib['cancel']
        except:
            pass 
    
    def returnModelObject(self):
        return self.modelObject



class DestVariables:
    def __init__(self):
        self.CALLSIGN = None

    @classmethod
    def geochat(cls):
        cls.CALLSIGN = None
        return cls

    @classmethod
    def other(cls):
        cls.CALLSIGN = None
        return cls


from FreeTAKServer.model.ExCheck.templateInstanceContents import templateInstanceContents
import datetime as dt

class templateInstance:
    def __init__(self):
        self.name = "exchecktemplates"
        self.description = ""
        self.chatRoom = ""
        self.tool = "ExCheck"
        self.keywords = []
        self.creatorUid = "ExCheck"
        self.createTime = ""
        self.externalData = []
        self.uids = []
        self.contents = []
        self.setcreateTime()

    def setcontents(self, content):
        if isinstance(content, templateInstanceContents):
            self.contents.append(content)
        else:
            raise Exception("invalid type passed to setcontents this setter only accepts templateInstanceContentsJson instances")

    def setname(self, name):
        self.name = name

    def getname(self):
        return self.name

    def setdescription(self, description):
        self.description = description

    def getdescription(self):
        return self.description

    def setcreateTime(self, createTime=None):
        DATETIME_FMT = "%Y-%m-%dT%H:%M:%S.%fZ"
        if createTime == None:
            timer = dt.datetime
            now = timer.utcnow()
            zulu = now.strftime(DATETIME_FMT)
            self.createTime = zulu
        else:
            self.createTime = createTime


class _GroupVariables:
    def __init__(self):
        self.NAME = None
        self.ROLE = None

    @classmethod
    def connection(cls):
        cls.NAME = None
        cls.ROLE = None
        return cls

    @classmethod
    def UserUpdate(cls):
        cls.name = None
        cls.role = None
        return cls

    @classmethod
    def Presence(cls):
        cls.name = None
        cls.role = None
        return cls


from FreeTAKServer.model.SpecificCoT.SendGeoChat import SendGeoChat
from .SendCoTAbstractController import SendCoTAbstractController
from FreeTAKServer.controllers.configuration.LoggingConstants import LoggingConstants
from FreeTAKServer.controllers.CreateLoggerController import CreateLoggerController

loggingConstants = LoggingConstants()
logger = CreateLoggerController("SendGeoChatController").getLogger()

class SendGeoChatController(SendCoTAbstractController):
    def __init__(self, RawCoT=None):
        try:
            tempObject = super().Event.GeoChat()
            object = SendGeoChat()
            if RawCoT is not None:
                self.fill_object(object, tempObject, RawCoT)
            else:
                pass
        except Exception as e:
            logger.error("there has been an exception in the creation"
                         " of the send Geo Chat object " + str(e))
            return -1


from FreeTAKServer.model.FTSModel.fts_protocol_object import FTSProtocolObject
from FreeTAKServer.model.FTSModelVariables.CreatorCallsignVariables import CreatorCallsignVariables as vars

class CreatorCallsign(FTSProtocolObject):
    @staticmethod
    def Checklist(INTAG = vars.Checklist().INTAG):
        creatorCallsign = CreatorCallsign()
        creatorCallsign.setINTAG(INTAG)
        return creatorCallsign

    def setINTAG(self, INTAG):
        self.INTAG = INTAG

    def getINTAG(self):
        return self.INTAG


from FreeTAKServer.controllers.MainSocketController import MainSocketController
import socket

class TCPSocketController(MainSocketController):
    def createSocket(self):
        self.MainSocket.sock = socket.socket(self.MainSocket.socketAF, self.MainSocket.socketSTREAM)
        self.MainSocket.sock.setsockopt(self.MainSocket.solSock, self.MainSocket.soReuseAddr, self.MainSocket.sockProto)
        self.MainSocket.sock.bind((self.MainSocket.ip, self.MainSocket.port))
        return self.MainSocket.sock


from lxml import etree
from FreeTAKServer.model.SpecificCoT.SendOther import SendOther
from .SendCoTAbstractController import SendCoTAbstractController
from FreeTAKServer.controllers.configuration.LoggingConstants import LoggingConstants
from FreeTAKServer.controllers.CreateLoggerController import CreateLoggerController

loggingConstants = LoggingConstants()
logger = CreateLoggerController("SendOtherController").getLogger()
class SendOtherController(SendCoTAbstractController):
    def __init__(self, RawCoT=None):
        if type(RawCoT != bytes):
            pass
        else:
            RawCoT.xmlString.encode()
        try:
            tempObject = super().Event.Other()
            self.object = SendOther()
            if RawCoT is not None:
                xml = RawCoT.xmlString
                RawCoT.xmlString = etree.tostring(self.filter_CoT(xml))
                self.fill_object(self.object, tempObject, RawCoT)
                try:
                    object = self.getObject()

                except Exception as e:
                    logger.error("there has been an exception getting object " + str(e))
                self.Object.setXmlString(xml)
            else:
                pass
        except Exception as e:
            logger.error("there has been an exception in the creation of an"
                         "Other object " + str(e))
    #this function modifies the CoT so only the marti and point tags are present
    def filter_CoT(self, event):
        try:
            outputEvent = etree.fromstring(event)
            tempDetail = outputEvent.find('detail')
            outputEvent.remove(tempDetail)
            detail = etree.fromstring(event).find('detail')
            try:
                marti = detail.find('marti')
                outputDetail = etree.SubElement(outputEvent, 'detail')
                outputDetail.append(marti)
                print('dest client found')
                self.object.martiPresent = True
            except:
                etree.SubElement(outputEvent, 'detail')
                print('no dest client found')

            return outputEvent
        except Exception as e:
            print('exception filtering CoT ' + str(e))


import multiprocessing
import threading
import argparse
from FreeTAKServer.controllers.CreateStartupFilesController import CreateStartupFilesController
from FreeTAKServer.controllers.services.TCPDataPackageService import TCPDataPackageService as TCPFlaskFunctions
from FreeTAKServer.controllers.services.SSLDataPackageService import SSLDataPackageService as SSLFlaskFunctions
from FreeTAKServer.controllers.configuration.OrchestratorConstants import OrchestratorConstants
from FreeTAKServer.controllers.configuration.LoggingConstants import LoggingConstants
from FreeTAKServer.controllers.CreateLoggerController import CreateLoggerController
from FreeTAKServer.controllers.services.RestAPI import RestAPI
from FreeTAKServer.model.ServiceObjects.FTS import FTS as FTSObj
from FreeTAKServer.model.SimpleClient import SimpleClient
import time
from FreeTAKServer.controllers.AddDataToCoTList import AddDataToCoTList
from FreeTAKServer.model.FilterGroup import FilterGroup
from FreeTAKServer.controllers.services.SSLCoTServiceController import SSLCoTServiceController
from FreeTAKServer.controllers.services.TCPCoTServiceController import TCPCoTServiceController
from FreeTAKServer.controllers.services.federation.FederationClientService import FederationClientServiceController
from FreeTAKServer.controllers.services.federation.federation import FederationServerService
from FreeTAKServer.controllers.services.FederationServerServiceController import FederationServerServiceController
from FreeTAKServer.controllers.DatabaseControllers.DatabaseController import DatabaseController
from FreeTAKServer.controllers.certificate_generation import AtakOfTheCerts

loggingConstants = LoggingConstants()
logger = CreateLoggerController("FTS").getLogger()


# noinspection PyUnresolvedReferences
class FTS:

    def __init__(self):
        self.CoTService = None
        self.TCPDataPackageService = None
        self.UserCommand = None
        self.killSwitch = False
        self.ReceiveConnectionsReset = None
        self.CoTPoisonPill = None
        self.ClientDataPipe = None
        self.SSLClientDataPipe = None
        self.clientArray = []
        self.socketCount = 0
        self.pipeList = {}
        self.FilterGroup = FilterGroup()
        self.FTSServiceStartupConfigObject = FTSObj()
        self.dbController = DatabaseController()
        logger.propagate = True
        logger.info('something')

    def start_restAPI_service(self, StartupObjects):
        try:
            RestAPIPipe, self.RestAPIPipe = multiprocessing.Pipe(duplex=True)
            self.RestAPICommands, self.RestAPICommandsFTS = multiprocessing.Pipe()
            self.receive_Rest_stopper = threading.Event()
            self.receive_Rest_stopper.clear()
            self.receive_Rest = threading.Thread(target=self.receive_Rest_commands, args=(self.receive_Rest_stopper,))
            self.RestAPIProcess = multiprocessing.Process(target=RestAPI().startup, args=(RestAPIPipe, self.RestAPICommands, StartupObjects.RestAPIService.RestAPIServiceIP, StartupObjects.RestAPIService.RestAPIServicePort, self.StartupTime))
            self.receive_Rest.start()
            self.RestAPIProcess.start()
            self.pipeList['restAPI'] = self.RestAPIPipe
            self.FilterGroup.sources.append(self.RestAPIPipe)
            return 1
        except Exception as e:
            logger.error('There has been an exception thrown in the startup of the restAPI service '+str(e))
            return -1

    def stop_RestAPI_service(self):
        try:
            self.RestAPIProcess.terminate()
            self.RestAPIProcess.join()
            self.receive_Rest_stopper.set()
            self.receive_Rest.join()
        except Exception as e:
            logger.error('an exception has been thrown in RestAPI Startup ' + str(e))

    def start_CoT_service(self, FTSServiceStartupConfigObject):
        try:
            self.ClientDataPipe, ClientDataPipeParentChild = multiprocessing.Pipe()
            self.TCPCoTService, self.TCPCoTServiceFTSPipe = multiprocessing.Pipe(duplex=True)
            self.CoTPoisonPill = multiprocessing.Event()
            self.CoTPoisonPill.set()
            self.ReceiveConnectionsReset = multiprocessing.Event()
            self.CoTService = multiprocessing.Process(target=TCPCoTServiceController().start, args=(FTSServiceStartupConfigObject.CoTService.CoTServiceIP, FTSServiceStartupConfigObject.CoTService.CoTServicePort, self.CoTPoisonPill, ClientDataPipeParentChild, self.ReceiveConnectionsReset, self.TCPCoTService))
            self.CoTService.start()
            self.pipeList['TCPCoTServiceFTSPipe'] = self.TCPCoTServiceFTSPipe
            self.FilterGroup.receivers.append(self.TCPCoTServiceFTSPipe)
            self.FilterGroup.sources.append(self.TCPCoTServiceFTSPipe)
            print('CoTService started')
            return 1
        except Exception as e:
            logger.error('an exception has been thrown in CoT service startup ' + str(e))
            return -1

    def stop_CoT_service(self):
        try:
            self.ClientDataPipe.close()
            self.CoTPoisonPill.clear()

            time.sleep(1)
            if self.CoTService.is_alive():
                self.CoTService.terminate()
                self.CoTService.join()
            else:
                self.CoTService.join()

            self.FilterGroup.sources.remove(self.TCPCoTServiceFTSPipe)
            self.FilterGroup.receivers.remove(self.TCPCoTServiceFTSPipe)

        except Exception as e:
            logger.error("there's been an exception in the stopping of CoT Service " + str(e))
            return -1
        return 1

    def start_tcp_data_package_service(self, FTSServiceStartupConfigObject):
        try:
            tcp_data_package_service_pipe, self.tcp_data_package_service_pipe = multiprocessing.Pipe(duplex=True)
            print('start 213')
            self.TCPDataPackageService = multiprocessing.Process(target=TCPFlaskFunctions().startup,
                                                                 args=(FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceIP, FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServicePort, tcp_data_package_service_pipe))
            print('starting now')
            self.TCPDataPackageService.start()
            self.pipeList['tcp_data_package_service_pipe'] = self.tcp_data_package_service_pipe
            self.FilterGroup.sources.append(self.tcp_data_package_service_pipe)
            time.sleep(2)
            return 1
        except Exception as e:
            logger.error('there has been an exception in the individual starting of the Data Packages Service ' + str(e))
            return -1

    def stop_tcp_data_package_service(self):
        del self.pipeList['tcp_data_package_service_pipe']
        self.FilterGroup.sources.remove(self.tcp_data_package_service_pipe)
        try:
            self.TCPDataPackageService.terminate()
        except Exception as e:
            logger.error("there's been an exception in the termination of DataPackage Service " + str(e))
            return -1
        try:
            self.TCPDataPackageService.join()
        except Exception as e:
            logger.error("there's been an exception in the joining of DataPackage Service " + str(e))
            return -1
        self.tcp_data_package_service_pipe.close()
        return 1

    def start_ssl_data_package_service(self, FTSServiceStartupConfigObject):
        try:
            print('start 213')
            ssl_data_package_service, self.ssl_data_package_service = multiprocessing.Pipe(duplex=True)
            self.SSLDataPackageService = multiprocessing.Process(target=SSLFlaskFunctions().startup,
                                                              args=(FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceIP, FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServicePort, ssl_data_package_service))
            print('starting SSL now')
            self.SSLDataPackageService.start()
            self.pipeList['ssl_data_package_service'] = self.ssl_data_package_service
            self.FilterGroup.sources.append(self.ssl_data_package_service)
            time.sleep(2)
            return 1
        except Exception as e:
            logger.error('there has been an exception in the individual starting of the Data Packages Service ' + str(e))
            return -1

    def stop_ssl_data_package_service(self):
        del(self.pipeList['ssl_data_package_service'])
        self.FilterGroup.sources.remove(self.ssl_data_package_service)
        try:
            self.SSLDataPackageService.terminate()
        except Exception as e:
            logger.error("there's been an exception in the termination of DataPackage Service " + str(e))
            return -1
        try:
            self.SSLDataPackageService.join()
        except Exception as e:
            logger.error("there's been an exception in the joining of DataPackage Service " + str(e))
            return -1
        self.ssl_data_package_service.close()
        return 1

    def start_SSL_CoT_service(self, FTSServiceStartupConfigObject):
        try:
            self.SSLClientDataPipe, SSLClientDataPipeParentChild = multiprocessing.Pipe()
            self.SSLCoTServicePipe, self.SSLCoTServiceFTSPipe = multiprocessing.Pipe()
            self.SSLCoTPoisonPill = multiprocessing.Event()
            self.SSLCoTPoisonPill.set()
            self.ReceiveConnectionsReset = multiprocessing.Event()
            self.SSLCoTService = multiprocessing.Process(target=SSLCoTServiceController().start, args=(FTSServiceStartupConfigObject.SSLCoTService.SSLCoTServiceIP, FTSServiceStartupConfigObject.SSLCoTService.SSLCoTServicePort, self.SSLCoTPoisonPill, SSLClientDataPipeParentChild, self.ReceiveConnectionsReset, self.SSLCoTServicePipe))
            self.SSLCoTService.start()
            self.pipeList['SSLCoTServiceFTSPipe'] = self.SSLCoTServiceFTSPipe
            self.FilterGroup.sources.append(self.SSLCoTServiceFTSPipe)
            self.FilterGroup.receivers.append(self.SSLCoTServiceFTSPipe)
            print('SSL CoTService started')
            return 1
        except Exception as e:
            logger.error('an exception has been thrown in SSL CoT service startup ' + str(e))
            return -1

    def stop_SSL_CoT_service(self):
        try:
            self.SSLClientDataPipe.close()
            self.SSLCoTPoisonPill.clear()

            time.sleep(0.1)
            if self.SSLCoTService.is_alive():
                self.SSLCoTService.terminate()
                self.SSLCoTService.join()
            else:
                self.SSLCoTService.join()
            del(self.pipeList["SSLCoTServiceFTSPipe"])
            self.FilterGroup.sources.remove(self.SSLCoTServicePipe)
            self.FilterGroup.receivers.remove(self.SSLCoTServicePipe)

        except Exception as e:
            logger.error("there's been an exception in the stopping of CoT Service " + str(e))
            return -1
        return 1

    def start_federation_client_service(self, FTSServiceStartupConfigObject):
        FederationClientServicePipe, self.FederationClientServicePipeFTS = multiprocessing.Pipe(True)
        self.FederationClientService = multiprocessing.Process(target=FederationClientServiceController().start, args=(FederationClientServicePipe,))
        self.FederationClientService.start()
        self.pipeList['FederationClientServiceFTSPipe'] = self.FederationClientServicePipeFTS
        self.FilterGroup.sources.append(self.FederationClientServicePipeFTS)
        self.FilterGroup.receivers.append(self.FederationClientServicePipeFTS)
        return 1

    def stop_federation_client_service(self):
        try:
            del(self.pipeList['FederationClientServiceFTSPipe'])
            self.FilterGroup.receivers.remove(self.FederationClientServicePipeFTS)
            self.FilterGroup.sources.remove(self.FederationClientServicePipeFTS)
            if self.FederationClientService.is_alive():
                self.FederationClientService.terminate()
                self.FederationClientService.join()
            else:
                self.FederationClientService.join()

            self.FilterGroup.sources.remove(self.FederationClientServicePipeFTS)
            self.FilterGroup.receivers.remove(self.FederationClientServicePipeFTS)
            return 1
        except:
            return -1
        
    def start_federation_server_service(self, FTSServiceStartupConfigObject):
            try:
                ip = FTSServiceStartupConfigObject.FederationServerService.FederationServerServiceIP
                port = FTSServiceStartupConfigObject.FederationServerService.FederationServerServicePort
                FederationServerServicePipe, self.FederationServerServicePipeFTS = multiprocessing.Pipe()
                self.FederationServerService = multiprocessing.Process(
                    target=FederationServerService().start, args=(FederationServerServicePipe, ip, port))
                self.FederationServerService.start()
                self.pipeList['FederationServerServiceFTSPipe'] = self.FederationServerServicePipeFTS
                self.FilterGroup.sources.append(self.FederationServerServicePipeFTS)
                self.FilterGroup.receivers.append(self.FederationServerServicePipeFTS)
                return 1
            except:
                return -1

    def stop_federation_server_service(self):
        try:
            if self.FederationServerService.is_alive():
                self.FederationServerService.terminate()
                self.FederationServerService.join()
            else:
                self.FederationServerService.join()
            self.FederationServerServicePipeFTS.close()
            self.FilterGroup.sources.remove(self.FederationServerServicePipeFTS)
            self.FilterGroup.receivers.remove(self.FederationServerServicePipeFTS)
            return 1
        except:
            return -1

    #change object name to FTSServiceStartupConfigObject
    def start_all(self, FTSServiceStartupConfigObject):
        import copy
        try:
            if FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceStatus == 'start':
                self.FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceStatus = FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceStatus
                self.start_tcp_data_package_service(FTSServiceStartupConfigObject)
                if FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServicePort != "":
                    self.FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServicePort = FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServicePort
                if FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceIP != "":
                    self.FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceIP = FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceIP
                
            elif FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceStatus == 'stop':
                self.FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceStatus = FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceStatus
                self.stop_tcp_data_package_service()
            else:
                if FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServicePort != "":
                    self.FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServicePort = FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServicePort
                if FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceIP != self.FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceIP:
                    if FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceStatus or self.FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceStatus:
                        self.FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceIP = FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceIP
                        self.stop_tcp_data_package_service()
                        self.start_tcp_data_package_service(FTSServiceStartupConfigObject)
                    else:
                        self.FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceIP = FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceIP

            if FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceStatus == 'start':
                self.FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceStatus = FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceStatus
                self.start_ssl_data_package_service(FTSServiceStartupConfigObject)
                if FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServicePort != "":
                    self.FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServicePort = FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServicePort
                if FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceIP != "":
                    self.FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceIP = FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceIP
            elif FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceStatus == 'stop':
                self.FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceStatus = FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceStatus
                self.stop_ssl_data_package_service()
            else:
                if FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServicePort != "":
                    self.FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServicePort = FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServicePort
                if FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceIP != self.FTSServiceStartupConfigObject.TCPDataPackageService.TCPDataPackageServiceIP:
                    if FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceStatus or self.FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceStatus:
                        self.FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceIP = FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceIP
                        self.stop_ssl_data_package_service()
                        self.start_ssl_data_package_service(FTSServiceStartupConfigObject)
                    else:
                        self.FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceIP = FTSServiceStartupConfigObject.SSLDataPackageService.SSLDataPackageServiceIP

            if FTSServiceStartupConfigObject.CoTService.CoTServiceStatus == 'start':
                self.FTSServiceStartupConfigObject.CoTService.CoTServiceStatus = FTSServiceStartupConfigObject.CoTService.CoTServiceStatus
                self.start_CoT_service(FTSServiceStartupConfigObject)
                if FTSServiceStartupConfigObject.CoTService.CoTServicePort != "":
                    self.FTSServiceStartupConfigObject.CoTService.CoTServicePort = FTSServiceStartupConfigObject.CoTService.CoTServicePort
                if FTSServiceStartupConfigObject.CoTService.CoTServiceIP != "":
                    self.FTSServiceStartupConfigObject.CoTService.CoTServiceIP = FTSServiceStartupConfigObject.CoTService.CoTServiceIP

            elif FTSServiceStartupConfigObject.CoTService.CoTServiceStatus == 'stop':
                self.FTSServiceStartupConfigObject.CoTService.CoTServiceStatus = FTSServiceStartupConfigObject.CoTService.CoTServiceStatus
                self.stop_CoT_service()
            else:
                pass

            if FTSServiceStartupConfigObject.SSLCoTService.SSLCoTServiceStatus == 'start':
                self.FTSServiceStartupConfigObject.SSLCoTService.SSLCoTServiceStatus = FTSServiceStartupConfigObject.SSLCoTService.SSLCoTServiceStatus
                self.start_SSL_CoT_service(FTSServiceStartupConfigObject)
                if FTSServiceStartupConfigObject.SSLCoTService.SSLCoTServicePort != "":
                    self.FTSServiceStartupConfigObject.SSLCoTService.SSLCoTServicePort = FTSServiceStartupConfigObject.SSLCoTService.SSLCoTServicePort
                if FTSServiceStartupConfigObject.SSLCoTService.SSLCoTServiceIP != "":
                    self.FTSServiceStartupConfigObject.SSLCoTService.SSLCoTServiceIP = FTSServiceStartupConfigObject.SSLCoTService.SSLCoTServiceIP
                    
            elif FTSServiceStartupConfigObject.SSLCoTService.SSLCoTServiceStatus == 'stop':
                self.FTSServiceStartupConfigObject.SSLCoTService.SSLCoTServiceStatus = FTSServiceStartupConfigObject.SSLCoTService.SSLCoTServiceStatus
                self.stop_SSL_CoT_service()

            else:
                pass

            if FTSServiceStartupConfigObject.FederationClientService.FederationClientServiceStatus == 'start':
                self.FTSServiceStartupConfigObject.FederationClientService.FederationClientServiceStatus = FTSServiceStartupConfigObject.FederationClientService.FederationClientServiceStatus
                self.start_federation_client_service(FTSServiceStartupConfigObject)
                if FTSServiceStartupConfigObject.FederationClientService.FederationClientServicePort != "":
                    self.FTSServiceStartupConfigObject.FederationClientService.FederationClientServicePort = FTSServiceStartupConfigObject.FederationClientService.FederationClientServicePort
                if FTSServiceStartupConfigObject.FederationClientService.FederationClientServiceIP != "":
                    self.FTSServiceStartupConfigObject.FederationClientService.FederationClientServiceIP = FTSServiceStartupConfigObject.FederationClientService.FederationClientServiceIP


            elif FTSServiceStartupConfigObject.FederationClientService.FederationClientServiceStatus == 'stop':
                self.FTSServiceStartupConfigObject.FederationClientService.FederationClientServiceStatus = FTSServiceStartupConfigObject.FederationClientService.FederationClientServiceStatus
                self.stop_federation_client_service()

            else:
                pass

            if FTSServiceStartupConfigObject.FederationServerService.FederationServerServiceStatus == 'start':
                self.FTSServiceStartupConfigObject.FederationServerService.FederationServerServiceStatus = FTSServiceStartupConfigObject.FederationServerService.FederationServerServiceStatus
                self.start_federation_server_service(FTSServiceStartupConfigObject)
                if FTSServiceStartupConfigObject.FederationServerService.FederationServerServicePort != "":
                    self.FTSServiceStartupConfigObject.FederationServerService.FederationServerServicePort = FTSServiceStartupConfigObject.FederationServerService.FederationServerServicePort
                if FTSServiceStartupConfigObject.FederationServerService.FederationServerServiceIP != "":
                    self.FTSServiceStartupConfigObject.FederationServerService.FederationServerServiceIP = FTSServiceStartupConfigObject.FederationServerService.FederationServerServiceIP

            elif FTSServiceStartupConfigObject.FederationServerService.FederationServerServiceStatus == 'stop' \
                    and self.FTSServiceStartupConfigObject.FederationServerService.FederationServerServiceStatus != 'stop':
                self.FTSServiceStartupConfigObject.FederationServerService.FederationServerServiceStatus = FTSServiceStartupConfigObject.FederationServerService.FederationServerServiceStatus
                self.stop_federation_server_service()

            else:
                pass

            return 1
        except Exception as e:
            logger.error('there has been an exception in FTS start_all ' + str(e))
            return -1

    def restart_receive_connection_process(self):
        self.ReceiveConnectionsReset.set()
        return 1

    def receive_data_froCoT_service_thread(self, pipe, clientArray):
        found = 0
        try:
            if pipe.poll(timeout=0.01):
                data = pipe.recv()
                self.socketCount = data[2]
                if data[0] == 'add':
                    clientArray.append(data[1])
                    self.RestAPIPipe.send(data[1])
                else:
                    for client in clientArray:
                        if client.ID == data[1].ID:
                            clientArray.remove(client)
                            found = 1
                        else:
                            pass
                    if found == 0:
                        for client in clientArray:
                            if client.IP == data[1].IP and client.modelObject.detail.contact.callsign == data[1].modelObject.detail.contact.callsign:
                                clientArray.remove(client)
                                found = 1
                            else:
                                pass
                    else:
                        pass

                return self.clientArray
            else:
                return self.clientArray
        except Exception as e:
            return self.clientArray

    def receive_Rest_commands(self, kill):
        while kill.is_set() == False:
            try:
                command = self.RestAPICommandsFTS.recv()
                if isinstance(command, tuple):
                    self.FederationClientServicePipeFTS.send(command)
                else:
                    function = getattr(self, command[0])
                    if len(command) >= 2:
                        output = function(command[1])
                    else:
                        output = function()
                    self.RestAPICommandsFTS.send(output)

            except Exception as e:
                logger.error('there has been an exception thrown in processing rest command ' + str(e))
                self.RestAPICommandsFTS.send('500')

    def check_server_status(self):
        return self.FTSServiceStartupConfigObject

    def help(self):
        print('start_all: to begin all services type')
        print('start_CoT_service: to begin CoT service type')
        print('start_data_package_service: to begin data package service  type')
        print('stop_all: to terminate all services type')
        print('stop_CoT_service: to terminate CoT service type')
        print('stop_data_package_service: to begin data package service type')
        print('check_service_status: to check the status of the services type')
        print('show_users: to show connected user information type')
        print('start_federation_client_service: to begin federation with another server')
        print('kill: to kill the full server type')

    def show_users(self):
        '''
        this method is used only by commands submitted over the rest interface
        '''
        data = [['', '', '']]
        objects = []
        output = []
        print(self.clientArray)
        for client in self.clientArray:
            data.append([client.IP, client.modelObject.detail.contact.callsign, client.modelObject.detail._group.name])
        for client in data:
            simpleClient = SimpleClient()
            simpleClient.callsign = client[1]
            simpleClient.team = client[2]
            simpleClient.ip = client[0]
            objects.append(simpleClient)

        output.append('total sockets: '+str(self.socketCount))
        print(output)
        return objects

    def verify_output(self, input, example=None):
        try:
            if example == None:
                if input == None or input == -1:
                    return False
                else:
                    return True

            else:
                if isinstance(input, example):
                    return True
                else:
                    return False
        except Exception as e:
            logger.error('there has been an exception in FTS verifying output ' + str(e))
            return False

    def stop_all(self):
        try:
            DataPackageServiceOutput = self.stop_data_package_service()
            if self.verify_output(DataPackageServiceOutput):
                pass
            else:
                raise Exception('error stopping DataPackage Service')
            CoTServiceOutput = self.stop_CoT_service()
            if self.verify_output(CoTServiceOutput):
                pass
            else:
                raise Exception('error stopping CoT Service')
            return 1
        except Exception as e:
            logger.error('there has been an exception in FTS stop_all ' + str(e))
            return -1

    def kill(self):
        try:
            self.killSwitch = True
            return 1
        except Exception as e:
            logger.error('error in kill function '+str(e))

    def checkPipes(self):
        try:
            for pipe in self.FilterGroup.sources:
                try:
                    data = AddDataToCoTList().recv(pipe)
                except Exception as e:
                    logger.error('get pipe data failed '+str(e))
                    continue
                #this runs in the event a new client has connected
                try:
                    if isinstance(data, list):
                        AddDataToCoTList().send(self.FilterGroup.receivers, data[0])
                        for client in self.clientArray:
                            AddDataToCoTList().send(self.FilterGroup.receivers, client)
                    # this runs in all other cases in which data is received
                    elif data != 0 and data is not None:
                        try:
                            print('data received in FTS ' + str(data.xmlString))
                        except:
                            pass
                        AddDataToCoTList().send(self.FilterGroup.receivers, data)
                    # this runs when a timeout is triggered
                    else:
                        pass
                except Exception as e:
                    logger.error('processing received connection data failed '+ str(e))
        except Exception as e:
            pass

    def startup(self, CoTPort, CoTIP, DataPackagePort, DataPackageIP, SSLDataPackagePort, SSLDataPackageIP, RestAPIPort, RestAPIIP, SSLCoTPort, SSLCoTIP, AutoStart, firstStart = False, UI="False"):
        try:
            self.dbController.remove_user()
            self.FTSServiceStartupConfigObject.RestAPIService.RestAPIServiceStatus = 'start'
            self.FTSServiceStartupConfigObject.RestAPIService.RestAPIServicePort = RestAPIPort
            self.FTSServiceStartupConfigObject.RestAPIServiceIP = RestAPIIP
            if firstStart:
                from datetime import datetime as dt
                self.StartupTime = dt.now()
            else:
                pass
            if AutoStart == 'False':
                StartupObject = FTSObj()
                StartupObject.RestAPIService.RestAPIServicePort = RestAPIPort
                StartupObject.RestAPIService.RestAPIServiceIP = RestAPIIP
                StartupObject.RestAPIService.RestAPIServiceStatus = 'start'
                self.start_restAPI_service(StartupObject)

            else:
                StartupObject = FTSObj()
                StartupObject.CoTService.CoTServiceIP = CoTIP
                StartupObject.CoTService.CoTServicePort = CoTPort
                StartupObject.CoTService.CoTServiceStatus = 'start'

                StartupObject.TCPDataPackageService.TCPDataPackageServiceIP = DataPackageIP
                StartupObject.TCPDataPackageService.TCPDataPackageServicePort = DataPackagePort
                StartupObject.TCPDataPackageService.TCPDataPackageServiceStatus = 'start'

                StartupObject.SSLDataPackageService.SSLDataPackageServiceIP = SSLDataPackageIP
                StartupObject.SSLDataPackageService.SSLDataPackageServicePort = SSLDataPackagePort
                StartupObject.SSLDataPackageService.SSLDataPackageServiceStatus = 'start'

                StartupObject.RestAPIService.RestAPIServicePort = RestAPIPort
                StartupObject.RestAPIService.RestAPIServiceIP = RestAPIIP
                StartupObject.RestAPIService.RestAPIServiceStatus = 'start'

                StartupObject.FederationClientService.FederationClientServiceStatus = 'start'

                #StartupObject.FederationServerService.FederationServerServiceStatus = ''

                StartupObject.SSLCoTService.SSLCoTServiceStatus = 'start'
                StartupObject.SSLCoTService.SSLCoTServiceIP = SSLCoTIP
                StartupObject.SSLCoTService.SSLCoTServicePort = SSLCoTPort
                self.start_restAPI_service(StartupObject)

                self.start_all(StartupObject)

            while True:
                try:
                    self.checkPipes()
                except Exception as e:
                    logger.error("error in core FTS process Data Pipe " + str(e))
                try:
                    self.clientArray = self.receive_data_froCoT_service_thread(self.ClientDataPipe, self.clientArray)
                except Exception as e:
                    logger.error("error thrown receiving clients from tcp CoT pipe " + str(e))
                try:
                    self.clientArray = self.receive_data_froCoT_service_thread(self.SSLClientDataPipe, self.clientArray)
                except Exception as e:
                    logger.error("error thrown receiving clients from SSL CoT pipe " + str(e))
        except Exception as e:
            logger.error('exception in the startup of FTS ' + str(e))

if __name__ == "__main__":
    """import importlib
    conf = importlib.import_module("FreeTAKServer-UI")
    FreeTAKServerUI = importlib.import_module("FreeTAKServer-UI.app", "run")
    y = FreeTAKServerUI.create_app()
    UIProc = multiprocessing.Process(target=FreeTAKServerUI.app, args=())
    UIProc.start()"""
    try:
        parser = argparse.ArgumentParser(description=OrchestratorConstants().FULLDESC)
        parser.add_argument('-CoTPort', type=int, help=OrchestratorConstants().COTPORTDESC,
                            default=FTSObj().CoTService.CoTServicePort)
        parser.add_argument('-CoTIP', type=str, help=OrchestratorConstants().COTIPDESC,
                            default=FTSObj().CoTService.CoTServiceIP)
        parser.add_argument('-SSLCoTPort', type=int, help=OrchestratorConstants().SSLCOTPORTDESC,
                            default=FTSObj().SSLCoTService.SSLCoTServicePort)
        parser.add_argument('-SSLCoTIP', type=str, help=OrchestratorConstants().SSLCOTIPDESC,
                            default=FTSObj().SSLCoTService.SSLCoTServiceIP)
        parser.add_argument('-DataPackagePort', type=int, help=OrchestratorConstants().APIPORTDESC,
                            default=FTSObj().TCPDataPackageService.TCPDataPackageServicePort)
        parser.add_argument('-DataPackageIP', type=str, help=OrchestratorConstants().APIPORTDESC,
                            default=FTSObj().TCPDataPackageService.TCPDataPackageServiceIP)
        parser.add_argument('-SSLDataPackagePort', type=int, help=OrchestratorConstants().APIPORTDESC,
                            default=FTSObj().SSLDataPackageService.SSLDataPackageServicePort)
        parser.add_argument('-SSLDataPackageIP', type=str, help=OrchestratorConstants().APIPORTDESC,
                            default=FTSObj().SSLDataPackageService.SSLDataPackageServiceIP)
        parser.add_argument('-RestAPIPort', type=int, help=OrchestratorConstants().APIPORTDESC,
                            default=FTSObj().RestAPIService.RestAPIServicePort)
        parser.add_argument('-RestAPIIP', type=str, help=OrchestratorConstants().APIPORTDESC,
                            default=FTSObj().RestAPIService.RestAPIServiceIP)
        parser.add_argument('-AutoStart', type=str, help='whether or not you want all services to start or only the root service and the RestAPI service', default='True')
        parser.add_argument('-UI', type=str, help="set to true if you would like to start UI on server startup")
        args = parser.parse_args()
        AtakOfTheCerts().bake_startup()
        CreateStartupFilesController()
        FTS().startup(args.CoTPort, args.CoTIP, args.DataPackagePort, args.DataPackageIP, args.SSLDataPackagePort, args.SSLDataPackageIP, args.RestAPIPort, args.RestAPIIP, args.SSLCoTPort, args.SSLCoTIP, args.AutoStart, True, args.UI)
    except Exception as e:
        print(e)


from .FTSVariables import FTSVariables as vars
from .CoTService import CoTService
from .TCPDataPackageService import TCPDataPackageService
from .SSLDataPackageService import SSLDataPackageService
from .RestAPIService import RestAPIService
from .SSLCoTService import SSLCoTService
from .FederationClientService import FederationClientService
from .FederationServerService import FederationServerService

class FTS:
    def __init__(self):
        self.CoTService = CoTService()
        self.TCPDataPackageService = TCPDataPackageService()
        self.SSLDataPackageService = SSLDataPackageService()
        self.RestAPIService = RestAPIService()
        self.SSLCoTService = SSLCoTService()
        self.FederationClientService = FederationClientService()
        self.FederationServerService = FederationServerService()


from FreeTAKServer.model.ExCheck.Checklists.checklistTask import checklistTask
class checklistTasks:
    def __init__(self):
        self.checklistTask = []
        self.__count = 0
    def setchecklistTask(self, checklistTaskobj):
        if isinstance(checklistTaskobj, checklistTask):
            self.checklistTask.append(checklistTaskobj)
        else:
            raise TypeError('unsupported type')

    def getchecklistTask(self):
        obj = self.checklistTask[self.__count]
        self.__count += 1
        return obj


from FreeTAKServer.model.FTSModel.fts_protocol_object import FTSProtocolObject
import datetime as dt
from FreeTAKServer.model.FTSModel.Detail import Detail
from .Point import Point
from FreeTAKServer.model.FTSModelVariables.EventVariables import EventVariables as vars
import uuid

class Event(FTSProtocolObject):
    # TODO: fix emergency methods
    # Event.py
    # Python implementation of the Class Event
    # represents a TAK event: this class is instantiated with a standard set of
    #    values.
    # Generated by Enterprise Architect
    # Created on: 11-Feb-2020 11:08:07 AM
    # Original author: Corvo
    #

    # event as an XML
    #<?xml version="1.0" encoding="UTF-8" standalone="yes"?><event version="2.0" uid="Linux-ABC.server-ping" type="b-t-f" time="2020-02-14T20:32:31.444Z" start="2020-02-14T20:32:31.444Z" stale="2020-02-15T20:32:31.444Z" how="h-g-i-g-o"> 
        
        #default constructor

    def __init__(self):
        
        
        self.version = None
        self.uid = None
        self.type = None
        self.how = None

        self.detail = None
        self.point = None
        self.start = None
        case = {

            'default': self.defaultFunc,

            'timeout': self.timeoutFunc
            
            }


        # flag to determin e if this event is a geo chcat if so, will be added as a
        # prefix to the uid
        
        # starting time when an event should be considered valid
        start = "%Y-%m-%dT%H:%M:%SZ"
        # basic event
        # Gives a hint about how the coordinates were generated
        

        # Schema version of this event instance (e.g.  2.0)
            
        # time stamp: when the event was generated
        time = "%Y-%m-%dT%H:%M:%SZ" 
        
        # Hierarchically organized hint about event type (defaultis is 'a-f-G-I'
        # for infrastructure)
        
            # ending time when an event should no longer be considered valid
        stale = "%Y-%m-%dT%H:%M:%SZ" 
        
            # Globally unique name for this information on this event can have
            # additional information attached.
        # e.g.  -ping means that this request is a ping
        
        # flag to determine if this event is a Ping, in this case append to the UID
        
    @staticmethod
    def Connection(version=vars.connection().VERSIONNUM, uid=vars.connection().UID, type=vars.connection().TYPE,
                  how=vars.connection().HOW, time=vars.connection().TIME, start=vars.connection().START,
                  stale=vars.connection().STALE):
        event = Event()
        event.setdetail(Detail.Connection())
        event.setpoint(Point())
        event.setversion(version)
        event.setuid(uid)
        event.settype(type)
        event.sethow(how)
        event.settime(time)
        event.setstart(start)
        event.setstale(stale)
        return event

    @staticmethod
    def GeoChat(version=vars.geochat().VERSIONNUM, uid=vars.geochat().UID, type=vars.geochat().TYPE,
                how=vars.geochat().HOW, time=vars.geochat().TIME, start=vars.geochat().START,
                stale=vars.geochat().STALE):
        event = Event()
        event.setdetail(Detail.GeoChat())
        event.setpoint(Point())
        event.setversion(version)
        event.setuid(uid)
        event.settype(type)
        event.sethow(how)
        event.settime(time)
        event.setstart(start)
        event.setstale(stale)
        return event

    @staticmethod
    def Ping(version=vars.ping().VERSIONNUM, uid=vars.ping().UID, type=vars.ping().TYPE,
             how=vars.ping().HOW, time=vars.ping().TIME, start=vars.ping().START,
             stale=vars.ping().STALE):
        event = Event()
        event.setdetail(Detail.Ping())
        event.setpoint(Point())
        event.setversion(version)
        event.setuid(uid)
        event.settype(type)
        event.sethow(how)
        event.settime(time)
        event.setstart(start)
        event.setstale(stale)
        return event

    @staticmethod
    def Other(version=vars.other().VERSIONNUM, uid=vars.other().UID, type=vars.other().TYPE,
              how=vars.other().HOW, time=vars.other().TIME, start=vars.other().START,
              stale=vars.other().STALE):
        event = Event()
        event.setdetail(Detail.Other())
        event.setpoint(Point())
        event.setversion(version)
        event.setuid(uid)
        event.settype(type)
        event.sethow(how)
        event.settime(time)
        event.setstart(start)
        event.setstale(stale)
        return event
               
    @staticmethod
    # this is the only constructor which doesn't instantiate detail as detail is a required protobuf
    # attribute and will therefore be in the message regardless
    def FederatedCoT(version=vars.FederatedCoT().VERSIONNUM, uid=vars.FederatedCoT().UID, type=vars.FederatedCoT().TYPE,
              how=vars.FederatedCoT().HOW, time=vars.FederatedCoT().TIME, start=vars.FederatedCoT().START,
              stale=vars.FederatedCoT().STALE):
        event = Event()
        event.setpoint(Point())
        event.setdetail(Detail.FederatedCoT())
        event.setversion(version)
        event.setuid(uid)
        event.settype(type)
        event.sethow(how)
        event.settime(time)
        event.setstart(start)
        event.setstale(stale)
        return event

    @staticmethod
    def emergecyOn(version=vars.emergency_on().VERSIONNUM, uid=vars.emergency_on().UID, type=vars.emergency_on().TYPE,
                   how=vars.emergency_on().HOW, time=vars.emergency_on().TIME, start=vars.emergency_on().START,
                   stale=vars.emergency_on().STALE):
        event = Event()
        event.setdetail(Detail.emergencyOn())
        event.setpoint(Point())
        event.setversion(version)
        event.setuid(uid)
        event.settype(type)
        event.sethow(how)
        event.settime(time)
        event.setstart(start)
        event.setstale(stale)
        return event

    @staticmethod
    def emergecyOff(version=vars.emergency_off().VERSIONNUM, uid=vars.emergency_off().UID, type=vars.emergency_off().TYPE,
                    how=vars.emergency_off().HOW, time=vars.emergency_off().TIME, start=vars.emergency_off().START,
                    stale=vars.emergency_off().STALE):
        event = Event()
        event.setdetail(Detail.emergencyOff())
        event.setpoint(Point())
        event.setversion(version)
        event.setuid(uid)
        event.settype(type)
        event.sethow(how)
        event.settime(time)
        event.setstart(start)
        event.setstale(stale)
        return event

    @staticmethod
    def dropPoint(version=vars.drop_point().VERSIONNUM, uid=vars.drop_point().UID, type=vars.drop_point().TYPE,
                  how=vars.drop_point().HOW, time=vars.drop_point().TIME, start=vars.drop_point().START,
                  stale=vars.drop_point().STALE):
        event = Event()
        event.setdetail(Detail.dropPoint())
        event.setpoint(Point())
        event.setversion(version)
        event.setuid(uid)
        event.settype(type)
        event.sethow(how)
        event.settime(time)
        event.setstart(start)
        event.setstale(stale)
        return event

    @staticmethod
    def disconnect(UID=vars.disconnect().uid, VERSION=vars.disconnect().version, TYPE=vars.disconnect().type,
                   TIME=vars.disconnect().time, START=vars.disconnect().start, STALE=vars.disconnect().stale,
                   HOW=vars.disconnect().how):
        event = Event()
        event.setuid(UID)
        event.setversion(VERSION)
        event.settype(TYPE)
        event.settime(TIME)
        event.setstart(START)
        event.setstale(STALE)
        event.sethow(HOW)
        event.point = Point()
        event.detail = Detail.disconnect()
        return event

    @staticmethod
    def takPong(VERSION=vars.takPong().version, UID=vars.takPong().uid, TYPE=vars.takPong().type,
                HOW=vars.takPong().how, TIME=vars.takPong().time, START=vars.takPong().start,
                STALE=vars.takPong().stale):
        event = Event()
        event.setversion(VERSION)
        event.setuid(UID)
        event.settype(TYPE)
        event.sethow(HOW)
        event.settime(TIME)
        event.setstart(START)
        event.setstale(STALE)
        event.point = Point()
        return event

    @staticmethod
    def UserUpdate(VERSION=vars.UserUpdate().version, UID=vars.UserUpdate().uid, TYPE=vars.UserUpdate().type,
                   HOW=vars.UserUpdate().how, TIME=vars.UserUpdate().time, START=vars.UserUpdate().start,
                   STALE=vars.UserUpdate().stale):
        event = Event()
        event.setversion(VERSION)
        event.setuid(UID)
        event.settype(TYPE)
        event.sethow(HOW)
        event.settime(TIME)
        event.setstart(START)
        event.setstale(STALE)
        event.point = Point()
        event.detail = Detail.UserUpdate()
        return event

    @staticmethod
    def SimpleCoT(VERSION=vars.SimpleCoT().version, UID=vars.SimpleCoT().uid, TYPE=vars.SimpleCoT().type,
                  TIME=vars.SimpleCoT().time, START=vars.SimpleCoT().start, STALE=vars.SimpleCoT().stale,
                  HOW=vars.SimpleCoT().how):
        event = Event()
        event.setversion(VERSION)
        event.setuid(UID)
        event.settype(TYPE)
        event.settime(TIME)
        event.setstart(START)
        event.setstale(STALE)
        event.sethow(HOW)
        event.detail = Detail.SimpleCoT()
        event.point = Point()
        return event

    @staticmethod
    def Presence(VERSION=vars.Presence().version, UID=vars.Presence().uid, TYPE=vars.Presence().type,
                  TIME=vars.Presence().time, START=vars.Presence().start, STALE=vars.Presence().stale,
                  HOW=vars.Presence().how):
        event = Event()
        event.setversion(VERSION)
        event.setuid(UID)
        event.settype(TYPE)
        event.settime(TIME)
        event.setstart(START)
        event.setstale(STALE)
        event.sethow(HOW)
        event.detail = Detail.Presence()
        event.point = Point()
        return event

    @staticmethod
    def ExcheckUpdate(VERSION=vars.ExcheckUpdate().version, UID=vars.ExcheckUpdate().uid,
                              TYPE=vars.ExcheckUpdate().type, TIME=vars.ExcheckUpdate().time,
                              START=vars.ExcheckUpdate().start, STALE=vars.ExcheckUpdate().stale,
                              HOW=vars.ExcheckUpdate().how):
        event = Event()
        event.setversion(VERSION)
        event.setuid(UID)
        event.settype(TYPE)
        event.settime(TIME)
        event.setstart(START)
        event.setstale(STALE)
        event.sethow(HOW)
        event.point = Point()
        event.detail = Detail.ExcheckUpdate()
        return event

    def defaultFunc(self, DATETIME_FMT,  version, uid, type, how, isGeochat, isPing):
        self.how = how

        timer = dt.datetime
        now = timer.utcnow()
        zulu = now.strftime(DATETIME_FMT)
        stale_part = dt.datetime.strptime(zulu, DATETIME_FMT) + dt.timedelta(minutes = 2)
        stale_part = stale_part.strftime(DATETIME_FMT)
        self.setstale(str(stale_part))
        self.setstart(zulu)
        self.settime(zulu)
        self.type = type
        self.setuid(isGeochat=isGeochat, isPing=isPing)
        self.version = version

    def timeoutFunc(self, DATETIME_FMT, version, uid, type, how, isGeochat, isPing):
        self.how = how

        timer = dt.datetime
        now = timer.utcnow()
        zulu = now.strftime(DATETIME_FMT)
        stale_part = dt.datetime.strptime(zulu, DATETIME_FMT) + dt.timedelta(minutes = 2)
        stale_part = stale_part.strftime(DATETIME_FMT)
        self.setstale(str(stale_part))
        self.setstart(zulu)
        self.settime(zulu)
        self.type = type
        self.setuid(isGeochat=isGeochat, isPing=isPing)
        self.version = version
        #start getter

    def getstart(self): 
        return self.start 
    
        # start setter
    def setstart(self, start=0):
        DATETIME_FMT = "%Y-%m-%dT%H:%M:%S.%fZ"
        if start == None:
            timer = dt.datetime
            now = timer.utcnow()
            zulu = now.strftime(DATETIME_FMT)
            self.start = zulu
        else:
            self.start = start

    
        # how getter
    def gethow(self): 
        return self.how 
    
        
    # how setter
    def sethow(self, how=0):  
        self.how = how 

        # uid getter
    def getuid(self): 
        return self.uid 
    
        # uid setter
    def setuid(self, uid):
        if uid == None:
            self.uid = str(uuid.uuid1())

        else:
            self.uid = uid

            # version getter
    def getversion(self): 
        return self.version 
    
        # version setter
    def setversion(self, version):  
        self.version = version 

            # time getter
    def gettime(self): 
        return self.time 
    
        # time setter
    def settime(self, time=0):
        DATETIME_FMT = "%Y-%m-%dT%H:%M:%S.%fZ"
        if time == None:
            timer = dt.datetime
            now = timer.utcnow()
            zulu = now.strftime(DATETIME_FMT)
            self.time = zulu
        else:
            self.time = time
        
        # stale getter
    def getstale(self): 
        return self.stale 
    
        # stale setter
    def setstale(self, stale = None,staletime=60):
        if stale == None:
            DATETIME_FMT = "%Y-%m-%dT%H:%M:%S.%fZ"
            timer = dt.datetime
            now = timer.utcnow()
            zulu = now.strftime(DATETIME_FMT)
            add = dt.timedelta(seconds=staletime)
            stale_part = dt.datetime.strptime(zulu, DATETIME_FMT) + add
            self.stale = stale_part.strftime(DATETIME_FMT)
        else:
            self.stale = stale

            # type getter
    def gettype(self): 
        return self.type 
    
        # type setter
    def settype(self, type=0):  
        self.type = type

    def getpoint(self):
        return self.point

        # type setter

    def setpoint(self, Point=None):
        self.point = Point

    def getdetail(self):
        return self.detail

        # type setter

    def setdetail(self, detail=None):
        self.detail = detail
if __name__ == "__main__":
    Event()




from FreeTAKServer.model.FTSModel.fts_protocol_object import FTSProtocolObject
#######################################################
# 
# __chat.py
# Python implementation of the Class __chat
# Generated by Enterprise Architect
# Created on(FTSProtocolObject):      11-Feb-2020 11(FTSProtocolObject):08(FTSProtocolObject):10 AM
# Original author: Corvo
# 
#######################################################
from FreeTAKServer.model.FTSModelVariables.ChatVariables import ChatVariables as vars
from FreeTAKServer.model.FTSModel.Chatgrp import Chatgrp

class Chat(FTSProtocolObject):
      # default constructor       
    def __init__(self):
        self.senderCallsign = None
        self.id = None
        self.parent = None
        self.chatroom = None
        self.groupOwner = None
        self.chatgrp = None

    @staticmethod
    def geochat(GROUPOWNER = vars.geochat().GROUPOWNER, ID = vars.geochat().ID, PARENT = vars.geochat().PARENT,
                CHATROOM = vars.geochat().CHATROOM, SENDERCALLSIGN = vars.geochat().SENDERCALLSIGN):
        chat = Chat()
        chat.setgroupOwner(GROUPOWNER)
        chat.setparent(PARENT)
        chat.setid(ID)
        chat.setchatroom(CHATROOM)
        chat.setsenderCallsign(SENDERCALLSIGN)
        chat.setchatgrp(Chatgrp.geochat())
        return chat

    def getparent(self): 
        return self.parent 

     # parent setter 
    def setparent(self,parent=0):  
        self.parent=parent 
    
    # senderCallsign getter 
    def getsenderCallsign(self):
        return senderCallsign 

    # senderCallsign setter 
    def setsenderCallsign(self,senderCallsignn):
        global senderCallsign
        senderCallsign=senderCallsignn 

      # chatroom getter 
    def getchatroom(self): 
        return self.chatroom 

    # chatroom setter 
    def setchatroom(self, chatroom=0):  
        self.chatroom=chatroom 

        # groupOwner getter 
    def getgroupOwner(self): 
        return self.groupOwner 

    # groupOwner setter 
    def setgroupOwner(self, groupOwner=0):  
        self.groupOwner=groupOwner 

      # id getter 
    def getid(self): 
        return self.id 

    # id setter 
    def setid(self, id=0):  
        self.id=id
    #chatgrp uid0 getter
    def getuid0(self):
      self.chatgrp().getuid0()
  
    def setuid0(self, uid0=0):
        self.chatgrp().setuid0(uid0)

    def getuid1(self):
        self.chatgrp().getuid1()
  
    def setuid1(self, uid1=0):
        self.chatgrp().setuid1(uid1)

    def getchatgrp(self):
        return self.chatgrp
  
    def setchatgrp(self, chatgrp=0):
        print(chatgrp)
        self.chatgrp = chatgrp



#######################################################
# 
# XMLCoTController.py
# Python implementation of the Class XMLCoTController
# Generated by Enterprise Architect
# Created on:      20-May-2020 1:07:38 PM
# Original author: Natha Paquette
# 
#######################################################
from lxml import etree


from FreeTAKServer.controllers.configuration.LoggingConstants import LoggingConstants
from FreeTAKServer.controllers.CreateLoggerController import CreateLoggerController
from FreeTAKServer.controllers.SpecificCoTControllers import *

logger = CreateLoggerController("XMLCoTController").getLogger()
loggingConstants = LoggingConstants()

class XMLCoTController:
    def __init__(self):  
        pass

    def determineCoTGeneral(self, data):
        # this will establish the CoTs general type
        if data.type == 'RawConnectionInformation':
            #this handels the event of a connection CoT
            try:
                return ("clientConnected", data)

            except Exception as e:
                logger.error(loggingConstants.XMLCOTCONTROLLERDETERMINECOTGENERALERRORA+str(e))
        #this runs if it is infact regular data
        elif data.xmlString == b'' or data.xmlString == None:
            #this handeles a client dissconection CoT
            return ("clientDisconnected", data)
        else:
            #this is the default in the event of an generic CoT or a CoT without a specific associated use case in the orchestrator
            try:
                return ("dataReceived", data)

            except Exception as e:
                logger.error(loggingConstants.XMLCOTCONTROLLERDETERMINECOTGENERALERRORB+str(e))

    def convert_model_to_row(self, modelObject, rowObject):
        for attribName, attribValue in modelObject.__dict__.items():
            if hasattr(attribValue, '__dict__'):
                subTableRow = getattr(rowObject, attribName)
                subTableRowObject = self.convert_model_to_row(attribValue, subTableRow)
                setattr(rowObject, attribName, subTableRowObject)
            else:
                setattr(rowObject, attribName, attribValue)

    def determineCoTType(self, RawCoT):
        # this function is to establish which specific controller applys to the CoT if any
        try:
            xml = RawCoT.xmlString
            if type(xml) != type(b''):
                xml = xml.encode()
            else:
                pass
            event = etree.fromstring(xml)
            detail = event.find('detail')
            CoTTypes = {
                            "*": "SendOtherController",
                            "emergency": "SendEmergencyController",
                            "invalid": "SendInvalidCoTController",
                            "health": "SendHealthCheckController",
                            "ping": "SendPingController",
                            "geochat": "SendGeoChatController",
                            "point": "SendDropPointController",
                            "userupdate": "SendUserUpdateController"
                            }
            # TODO: the below if statement is probably unnecessary but this needs to be verified
            if RawCoT == b'' or RawCoT == None:
                RawCoT.disconnect = 1

            elif detail.find('emergency') != None:
                RawCoT.CoTType = CoTTypes['emergency']
                emergency = detail.find('emergency')
                try:
                    if emergency.attrib['cancel'] == 'true':
                        RawCoT.status = 'off'
                except:
                    RawCoT.status = 'on'

            elif str(event.attrib['type']) == "t-x-c-t":
                RawCoT.CoTType = CoTTypes['ping']
                return RawCoT

            elif str(event.attrib['type']) == "b-t-f":
                RawCoT.CoTType = CoTTypes['geochat']
                return RawCoT

            elif str(event.attrib['type']) in ["a-f-G-U-C", "a-f-G-E-V-A", "a-f-G-U-C-I", "a-f-G-E-V-C", "a-f-G-U", "a-f-G-E-V-A"]:
                RawCoT.CoTType = CoTTypes['userupdate']
                return RawCoT

            elif str(event.attrib['type']) == "a-h-G" or str(event.attrib['type']) == "a-n-G" or str(event.attrib['type']) == "a-f-G" or str(event.attrib['type']) == "a-u-G":
                RawCoT.CoTType = CoTTypes['point']
                return RawCoT
            
            elif str(event.attrib['type']) == "t-x-m-c":
                logger.debug('a txmc type xml has been received \n')
                RawCoT.CoTType = CoTTypes['*']
                return RawCoT 
            # TODO: this needs to be expanded for more use cases
            else:
                RawCoT.CoTType = CoTTypes['*']

            return RawCoT
        except Exception as e:
            RawCoT.CoTType = "SendInvalidCoTController"
            return RawCoT

    def categorize_type(self, type):
        from FreeTAKServer.controllers.RestMessageControllers.SendEmergencyController import SendEmergencyController
        from FreeTAKServer.controllers.SpecificCoTControllers.SendDropPointController import SendDropPointController
        from FreeTAKServer.controllers.SpecificCoTControllers.SendGeoChatController import SendGeoChatController
        from FreeTAKServer.controllers.SpecificCoTControllers.SendHealthCheckController import SendHealthCheckController
        from FreeTAKServer.controllers.SpecificCoTControllers.SendInvalidCoTController import SendInvalidCoTController
        from FreeTAKServer.controllers.SpecificCoTControllers.SendOtherController import SendOtherController
        from FreeTAKServer.controllers.SpecificCoTControllers.SendPingController import SendPingController
        from FreeTAKServer.controllers.SpecificCoTControllers.SendUserUpdateController import SendUserUpdateController
        CoTTypes = {
            "*": SendOtherController,
            "emergency": SendEmergencyController,
            "invalid": SendInvalidCoTController,
            "health": SendHealthCheckController,
            "ping": SendPingController,
            "geochat": SendGeoChatController,
            "point": SendDropPointController,
            "userupdate": SendUserUpdateController
        }
        if type == "t-x-c-t":
            return CoTTypes['ping']


        elif type == "b-t-f":
            return CoTTypes['geochat']


        elif type in ["a-f-G-U-C", "a-f-G-E-V-A", "a-f-G-U-C-I", "a-f-G-E-V-C", "a-f-G-U", "a-f-G-E-V-A"]:
            return CoTTypes['userupdate']


        elif type == "a-h-G" or type == "a-n-G" or type == "a-f-G" or type == "a-u-G":
            return CoTTypes['point']


        elif type == "t-x-m-c":
            logger.debug('a txmc type xml has been received \n')
            return CoTTypes['*']

            # TODO: this needs to be expanded for more use cases

        else:
            return CoTTypes['*']


    def findCallsign(self):
        pass

    def findMarti(self):
        pass

    def findUID(self):
        pass

    def serialize_model_to_CoT(self, modelObject, tagName = 'event', level = 0):
        xml = etree.Element(tagName)
        for attribName, value in modelObject.__dict__.items():
            if hasattr(value, '__dict__'):
                tagElement = self.serialize_model_to_CoT(value, attribName, level = level + 1)
                # TODO: modify so double underscores are handled differently
                try:
                    if attribName[0] == '_':
                        tagElement.tag = '_' + tagElement.tag
                        xml.append(tagElement)
                except:
                    pass
                else:
                    xml.append(tagElement)

            elif value == None:
                continue

            elif isinstance(value, list):
                for element in value:
                    tagElement = self.serialize_model_to_CoT(element, attribName, level=level + 1)
                    # TODO: modify so double underscores are handled differently
                    try:
                        if attribName[0] == '_':
                            tagElement.tag = '_' + tagElement.tag
                            xml.append(tagElement)
                    except:
                        pass
                    else:
                        xml.append(tagElement)

            # handles text data within tag
            elif attribName == "INTAG":
                xml.text = value

            else:
                # TODO: modify so double underscores are handled differently
                # handles instances in which attribute name begins with double underscore
                try:
                    if attribName[0] == '_':
                        xml.attrib['_'+attribName] = value
                except:
                    pass
                else:
                    xml.attrib[attribName] = str(value)

        if level == 0:
            return etree.tostring(xml)
        else:
            return xml

    def serialize_CoT_to_model(self, model, xml):
        attributes = xml.attrib
        if xml.text != None:
            setter = getattr(model, 'setINTAG')
            setter(xml.text)
        else:
            pass

        for key, value in attributes.items():
            setter = getattr(model, 'set'+key)
            setter(value)

        for element in xml:
            submodel = getattr(model, 'get'+element.tag)
            submodel = submodel()
            out = self.serialize_CoT_to_model(submodel, element)
            setter = getattr(model, 'set'+element.tag)
            setter(out)

        return model




from FreeTAKServer.controllers.configuration.LoggingConstants import LoggingConstants
from FreeTAKServer.controllers.CreateLoggerController import CreateLoggerController
loggingConstants = LoggingConstants()
logger = CreateLoggerController("SendDataController").getLogger()
import copy
#TODO: the part handling new connection from seperate process needs to be cleaned up

class SendDataController:

    def __init__(self):
        pass
    def sendDataInQueue(self, sender, processedCoT, clientInformationQueue, shareDataPipe = None):
        try:
            print('sending data to fts client' + str(processedCoT.xmlString))
        except Exception as e:
            print(e)
        try:
            if processedCoT.type == 'GeoChat':
                self.returnData = self.geochat_sending(clientInformationQueue, processedCoT, sender, shareDataPipe)
                return self.returnData
            elif sender == processedCoT:
                for client in clientInformationQueue:
                    try:
                        sock = client.socket
                        sock.send(processedCoT.idData.encode())
                        sender.socket.send(client.idData.encode())
                        # this is a special case which is identified
                        # by the server due to the list contents
                        # being within a list
                    except Exception as e:
                        logger.error('error in sending connection data ' + str(processedCoT.idData))
                        pass
                copiedProcessedCoTObject = copy.deepcopy(processedCoT)
                copiedProcessedCoTObject.idData = copiedProcessedCoTObject.idData.encode()
                shareDataPipe.send([copiedProcessedCoTObject])
                return 1
            elif processedCoT.type == 'other':
                self.returnData = self.send_to_specific_client(clientInformationQueue, processedCoT, sender, shareDataPipe)
                return self.returnData
            else:
                self.returnData = self.send_to_all(clientInformationQueue, processedCoT, sender, shareDataPipe)
                return self.returnData
        except Exception as e:
            logger.error(loggingConstants.SENDDATACONTROLLERSENDDATAINQUEUEERROR+str(e))
            return -1

    def send_to_specific_client(self, clientInformationQueue, processedCoT, sender, shareDataPipe):
        try:
            if processedCoT.martiPresent == False:
                print('marti not present')
                return self.send_to_all(clientInformationQueue, processedCoT, sender, shareDataPipe)

            else:
                print('marti present')
                for dest in processedCoT.modelObject.detail.marti.dest:
                    try:
                        for client in clientInformationQueue:
                            if client.modelObject.detail.contact.callsign == dest.callsign:
                                print('client socket is ' + str(client.socket))
                                sock = client.socket
                                try:
                                    sock.send(processedCoT.xmlString)
                                except Exception as e:
                                    logger.error('error sending data with marti to client data ' + str(
                                        processedCoT.xmlString) + 'error is ' + str(e))
                                    return (-1, client)
                            else:
                                continue
                    except Exception as e:
                        logger.error('error sending data with marti to client within if data is ' + str(
                            processedCoT.xmlString) + 'error is ' + str(e))
                        return -1
                if shareDataPipe != None:
                    processedCoT.clientInformation = None
                    shareDataPipe.send(processedCoT)
                else:
                    pass
                return 1
        except Exception as e:
            logger.error('error in send data to specific client ' + str(e))
            return -1
    def send_to_all(self, clientInformationQueue, processedCoT, sender, shareDataPipe):
        try:
            for client in clientInformationQueue:
    
                sock = client.socket
                try:
                    if hasattr(processedCoT, 'xmlString'):
                        print('sending to all ' + str(processedCoT.xmlString))
                        sock.send(processedCoT.xmlString)
                    else:
                        sock.send(processedCoT.idData.encode())
                except Exception as e:
                    print(e)
                    logger.error('error in sending of data ' + str(processedCoT.xmlString))
                    return (-1, client)
            if shareDataPipe != None:
                processedCoT.clientInformation = None
                shareDataPipe.send(processedCoT)
            else:
                pass
            return 1
        except Exception as e:
            logger.error('error in send to all ' + str(e))
    def geochat_sending(self, clientInformationQueue, processedCoT, sender, shareDataPipe):
        try:
            if processedCoT.modelObject.detail._chat.chatgrp.uid1 == 'All Chat Rooms':
                return self.send_to_all(clientInformationQueue, processedCoT, sender, shareDataPipe)
    
            else:
                for client in clientInformationQueue:
                    try:
                        if client.modelObject.uid == processedCoT.modelObject.detail._chat.chatgrp.uid1:
                            sock = client.socket
                            try:
                                sock.send(processedCoT.xmlString)
                            except Exception as e:
                                logger.error('error sending data with marti to client data ' + str(
                                    processedCoT.xmlString) + 'error is ' + str(e))
                                return (-1, client)
                        else:
                            continue
                    except Exception as e:
                        logger.error('error sending data with marti to client within if data is ' + str(
                            processedCoT.xmlString) + 'error is ' + str(e))
                        return -1
                if shareDataPipe != None:
                    processedCoT.clientInformation = None
                    shareDataPipe.send(processedCoT)
                else:
                    pass
                return 1
        except Exception as e:
            logger.error('there has been an exception in sending geochat ' + str(e))


#######################################################
# 
# serverdestination.py
# Python implementation of the Class serverdestination
# Generated by Enterprise Architect
# Created on:      28-Sep-2020 10:48:18 PM
# Original author: natha
# 
#######################################################
from sqlalchemy import Column, ForeignKey
from FreeTAKServer.model.SQLAlchemy.Root import Base
from sqlalchemy import String
from sqlalchemy.orm import relationship


class Serverdestination(Base):
# default constructor  def __init__(self):  
    __tablename__ = "Serverdestination"
    PrimaryKey = Column(ForeignKey("Detail.PrimaryKey"), primary_key=True)
    Detail = relationship("Detail", back_populates="serverdestination")
    # string composed by IP:port: protocol:machineID.
    # <i> e.g. 192.168.0.103:4242:tcp:ANDROID-R52JB0CDC4E</i>
    destinations = Column(String)


from FreeTAKServer.controllers.DatabaseControllers.table_controllers import TableController
from FreeTAKServer.model.SQLAlchemy.APIUsers import APIUsers

class APIUsersController(TableController):

    def __init__(self):
        self.table = APIUsers


from FreeTAKServer.controllers.configuration.MainConfig import MainConfig

class SSLDataPackageVariables:
    def __init__(self):
        self.SSLDataPackageIP = MainConfig.DataPackageServiceDefaultIP
        self.SSLDataPackagePort = 8443
        self.SSLDataPackageStatus = ""


class CreatorUidVariables:
    @classmethod
    def ExcheckUpdate(cls):
        cls.INTAG = None
        return cls

    @classmethod
    def Checklist(cls):
        cls.INTAG = None
        return cls




from typing import List

from FreeTAKServer.controllers.serializers.serializer_abstract import SerializerAbstract
from FreeTAKServer.model.FTSModel.Event import Event

class JsonSerializer(SerializerAbstract):

    def from_format_to_fts_object(self, object: dict, FTSObject: Event) -> Event:
        for key, value in object.items():
            if type(value) is dict:
                self._handle_sub_dictionary(FTSObject, key, value)
            else:
                self._handle_attribute(FTSObject, key, value)
        return FTSObject

    def _handle_attribute(self, FTSObject, key, value, expected_class_name = None):
        # retrieve setter list
        setters = self._get_fts_object_var_setter(FTSObject, key)
        # get correct setter
        setter = self._get_method_in_method_list(setters, expected_class_name)
        # call setter
        setter(value)

    def _handle_sub_dictionary(self, FTSObject, name: str, value: dict):
        for key, value in value.items():
            if value is dict:
                self._handle_sub_dictionary(FTSObject, key, value)
            else:
                self._handle_attribute(FTSObject, key, value, name)

    def _get_method_in_method_list(self, method_list: List[callable], expected_class_name: str = None) -> callable:
        if len(method_list) == 1:
            return method_list[0]

        elif len(method_list) == 0:
            raise AttributeError(expected_class_name + ' does not have specified attribute')

        elif len(method_list) > 1 and expected_class_name is not None:
            for method in method_list:
                if method.__self__.__class__.__name__.lower() == expected_class_name.lower():
                    return method
                else:
                    pass
            raise AttributeError('invalid json')

        else:
            raise AttributeError('invalid json')

    def from_fts_object_to_format(self, FTSObject: Event) -> dict:
        for key, value in object.items():
            getters = self._get_fts_object_var_getter(FTSObject, key)
            getter = self._get_method_in_method_list(getters, key)
            value = getter()
        return object_body
if __name__ == "__main__":
    getter = JsonSerializer().get_fts_object_var_getter(Event.Presence(), 'uid')
    print(getter())


from FreeTAKServer.controllers.configuration.MainConfig import MainConfig
from FreeTAKServer.controllers.configuration.types import Types
from FreeTAKServer.controllers.services.federation.handlers import StopHandler, DisconnectHandler, ConnectHandler, SendDataHandler, SendConnectionDataHandler, SendDisconnectionDataHandler
from FreeTAKServer.model.protobufModel.fig_pb2 import FederatedEvent
from FreeTAKServer.controllers.services.service_abstracts import ServerServiceInterface, ServiceBase
from multiprocessing import Pipe as multiprocessingPipe
from FreeTAKServer.model.federate import Federate
import selectors
import socket
from typing import Tuple
import ssl
import time
import codecs
from FreeTAKServer.controllers.serializers.protobuf_serializer import ProtobufSerializer
from FreeTAKServer.controllers.serializers.xml_serializer import XmlSerializer
from FreeTAKServer.controllers.XMLCoTController import XMLCoTController
from FreeTAKServer.model.SpecificCoT.SendOther import SendOther
from FreeTAKServer.model.FTSModel.Event import Event
from FreeTAKServer.controllers.DatabaseControllers.DatabaseController import DatabaseController
from FreeTAKServer.model.ClientInformation import ClientInformation
from FreeTAKServer.model.SpecificCoT.SendDisconnect import SendDisconnect
from FreeTAKServer.controllers.CreateLoggerController import CreateLoggerController
logger = CreateLoggerController("FederationServer").getLogger()


class FederationServerService(ServerServiceInterface, ServiceBase):

    def __init__(self):
        self._define_responsibility_chain()
        self.pipe = None
        self.federates: {str: Federate} = {}
        self.sel = selectors.DefaultSelector()

    def _send_connected_clients(self, connection):
        try:
            clients = self.db.query_user()
        except Exception as e:
            logger.warning("error thrown in getting clients from DataBase to send to federates " + str(e))
            return None
        for client in clients:
            try:
                proto_obj = FederatedEvent()
                proto_obj.contact.uid = str(client.uid)
                proto_obj.contact.callsign = str(client.CoT.detail.contact.callsign)
                proto_obj.contact.operation = 1
                proto_str = proto_obj.SerializeToString()
                header = self._generate_header(len(proto_str))
                connection.send(header+proto_str)
            except Exception as e:
                logger.warning("error thrown sending federate data to newly connected federate " + str(e))
                continue

    def _create_context(self) ->None:
        self.context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        self.context.load_cert_chain(MainConfig.federationCert, MainConfig.federationKey,
                                     password=MainConfig.federationKeyPassword)

    def _create_listener(self, ip: str, port: int)->None:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        sock.bind((ip, port))
        sock.listen()
        ssock = self.context.wrap_socket(sock, server_side=True)
        ssock.setblocking(False)
        self.sel.register(ssock, selectors.EVENT_READ, data=None)

    def _define_responsibility_chain(self):
        self.m_StopHandler = StopHandler()

        self.m_DisconnectHandler = DisconnectHandler()
        self.m_DisconnectHandler.setNextHandler(self.m_StopHandler)

        self.m_SendDataHandler = SendDataHandler()
        self.m_SendDataHandler.setNextHandler(self.m_DisconnectHandler)

        self.m_SendDisconnectionHandler = SendDisconnectionDataHandler()
        self.m_SendDisconnectionHandler.setNextHandler(self.m_SendDataHandler)

        # first handler in chain of responsibility and should be called first
        self.m_SendConnectionHandler = SendConnectionDataHandler()
        self.m_SendConnectionHandler.setNextHandler(self.m_SendDisconnectionHandler)

    def main(self):
        from lxml import etree
        while True:
            time.sleep(0.1)
            command = self.receive_command_data(self.pipe)
            if command:
                try:
                    self.m_SendConnectionHandler.Handle(self, command)
                except Exception as e:
                    pass
            else:
                pass
            try:
                data = self.receive_data_from_federate(1)
            except ssl.SSLWantReadError:
                data = None
            if data:
                for protobuf_object in data:
                    # TODO: clean all of this up as it's just a PoC

                    # event = etree.Element('event')
                    # SpecificCoTObj = XMLCoTController().categorize_type(protobuf_object.type)
                    try:
                        detail = etree.fromstring(protobuf_object.event.other)
                        protobuf_object.event.other = ''
                        fts_obj = ProtobufSerializer().from_format_to_fts_object(protobuf_object, Event.FederatedCoT())
                        specific_obj = SendOther()
                        event = XmlSerializer().from_fts_object_to_format(fts_obj)
                        xmlstring = event
                        xmlstring.find('detail').remove(xmlstring.find('detail').find('remarks'))
                        xmlstring.find('detail').extend([child for child in detail])
                        # specific_obj.xmlString = etree.tostring(xmlstring)
                        print(etree.tostring(xmlstring))
                        specific_obj.xmlString = etree.tostring(xmlstring)
                        self.pipe.send(specific_obj)
                    except Exception as e:
                        pass
                    """if isinstance(SpecificCoTObj, SendOtherController):
                        detail = protobuf_object.event.other
                        protobuf_object.event.other = ''
                        fts_obj = ProtobufSerializer().from_format_to_fts_object(protobuf_object, Event.Other())
                        protobuf_object.event.other = detail
                        SpecificCoTObj.object = fts_obj
                        SpecificCoTObj.Object =
                    else:
                        fts_obj = ProtobufSerializer().from_format_to_fts_object(protobuf_object, SpecificCoTObj().object)
                        self.pipe.send(data)"""
            else:
                pass

    def receive_data_from_federate(self, timeout):
        """called whenever data is available from any federate and immediately proceeds to
        send data through process pipe
        """
        dataarray = []
        events = self.sel.select(timeout)
        for key, mask in events:
            if key.data is None:
                self._accept_connection(key.fileobj)
            else:
                federate_data = self._receive_new_data(key)
                if federate_data:
                    dataarray.append(federate_data)
        return dataarray

    def _receive_new_data(self, key):
        try:
            conn = key.fileobj
            header = conn.recv(4)
            if header:
                try:
                    buffer = self._get_header_length(header)
                    raw_protobuf_message = conn.recv(buffer)
                    print(raw_protobuf_message)
                    protobuf_object = FederatedEvent()
                    protobuf_object.ParseFromString(raw_protobuf_message)
                    return protobuf_object
                except Exception as e:
                    conn.recv(10000)
                    return None
            else:
                self.disconnect_client(key.data.uid)
        except OSError:
            return None
        except Exception as e:
            logger.warning("exception in receiving data from federate "+str(e))
            self.disconnect_client(key.data.uid)
    
    def _accept_connection(self, sock) -> None:
        try:
            import uuid
            conn, addr = sock.accept()  # Should be ready to read
            print('accepted connection from', addr)
            conn.setblocking(False)
            data = Federate()
            data.conn = conn
            # get federate certificate CN
            # data.name = dict(x[0] for x in conn.getpeercert()["subject"])["commonName"]
            data.name = addr[0]
            data.addr = addr
            data.uid = str(uuid.uuid4())
            events = selectors.EVENT_READ | selectors.EVENT_WRITE
            self._send_connected_clients(conn)
            self.sel.register(conn, events, data=data)
            self.federates[data.uid] = data

            self.db.create_ActiveFederation(id = data.uid, federate = "unknown", address = addr[0], port = addr[1], initiator = "Remote")
            return None
        except Exception as e:
            print(e)
            logger.warning("exception thrown accepting federation " + str(e))

    def _get_header_length(self, header):
        return int.from_bytes(header, 'big')

    def _generate_header(self, contentlength):
        return contentlength.to_bytes(4, byteorder="big")

    def send_data_to_clients(self, data):
        from lxml import etree
        try:
            if self.federates:
                xmlstring = data.xmlString
                detail = etree.fromstring(xmlstring).find('detail')
                protobuf = ProtobufSerializer().from_fts_object_to_format(data.modelObject)
                protobuf.event.other = etree.tostring(detail)
                protobufstring = protobuf.SerializeToString()
                header = self._generate_header(len(protobufstring))
                protobufstring = header + protobufstring
                print(protobufstring)
                for client in self.federates.values():
                    client.conn.send(protobufstring)
            else:
                return None
        except Exception as e:
            logger.warning("send data to clients failed with exception "+str(e))

    def disconnect_client(self, id: str) -> None:
        try:
            logger.info("disconnecting client")
            try:
                federate = self.federates[id]
            except Exception as e:
                logger.warning("federate array has no item with uid " + str(id) + " federates array is len " + str(
                    len(self.federates)))
                return None
            try:
                federate.conn.close()
                self.sel.unregister(federate.conn)
                del (self.federates[federate.uid])
            except Exception as e:
                logger.warning("exception thrown disconnecting client " + str(e))

            try:
                self.db.remove_ActiveFederation(f'id == "{federate.uid}"')
            except Exception as e:
                logger.warning("exception thrown removing outgoing federation from DB " + str(e))
            return None
        except Exception as e:
            logger.warning("exception thrown accessing client for disconnecting client " + str(e))

    def send_connection_data(self, CoT: ClientInformation):
        if self.federates:
            proto_obj = FederatedEvent()
            proto_obj.contact.uid = str(CoT.modelObject.uid)
            proto_obj.contact.callsign = str(CoT.modelObject.detail.contact.callsign)
            proto_obj.contact.operation = 1
            proto_str = proto_obj.SerializeToString()
            header = self._generate_header(len(proto_str))
            for fed in self.federates.values():
                fed.conn.send(header + proto_str)
        else:
            return None

    def send_disconnection_data(self, CoT: SendDisconnect):
        if self.federates:
            proto_obj = FederatedEvent()
            proto_obj.contact.uid = str(CoT.modelObject.detail.link.uid)
            proto_obj.contact.callsign = str(CoT.modelObject.detail.link.type)
            proto_obj.contact.operation = 4
            proto_str = proto_obj.SerializeToString()
            header = self._generate_header(len(proto_str))
            for fed in self.federates.values():
                fed.conn.send(header + proto_str)
        else:
            return None

    def start(self, pipe, ip, port):
        self.db = DatabaseController()
        self.pipe = pipe
        self._create_context()
        self._create_listener(ip, port)
        print('started federation server service')
        self.main()

    def stop(self):
        pass

if __name__ == "__main__":

    from multiprocessing import Pipe
    pipe1, pipe2 = Pipe(True)
    FederationServerService().start(pipe1, "0.0.0.0", 9000)



#######################################################
# 
# Presence.py
# Python implementation of the Class Presence
# Generated by Enterprise Architect
# Created on:      04-Nov-2020 4:27:12 PM
# Original author: natha
# 
#######################################################
from FreeTAKServer.model.RestMessages.RestEnumerations import RestEnumerations


class Presence:
# default constructor  def __init__(self):  
    how = ""
    # the angular distance of the geoobject from the meridian of the greenwich, UK
    # expressed in positive or negative float. (e.g -76.107.7998)
    latitude = ""
    longitude = ""
    name = ""
    # the given role within the team . Please see API documentation for a list of
    # valid entries.
    role = ""
    # the color of the team
    team = ""
    timeout = ""


from FreeTAKServer.model.sockets.MainSocket import MainSocket
from FreeTAKServer.controllers.configuration.MainConfig import MainConfig

class SSLServerSocket(MainSocket):
    def __init__(self):
        super().__init__()
        self.keyDir = MainConfig.keyDir
        self.pemDir = MainConfig.pemDir
        self.password = MainConfig.password
        self.CA = MainConfig.CA


class TakvVariables:
    def __init__(self):
        self.VERSION = None
        self.PLATFORM = None
        self.DEVICE = None
        self.OS = None

    @classmethod
    def connection(cls):
        cls.VERSION = None
        cls.PLATFORM = None
        cls.DEVICE = None
        cls.OS = None
        return cls

    @classmethod
    def UserUpdate(cls):
        cls.device = None
        cls.platform = None
        cls.os = None
        cls.version = None
        return cls


class SQLcommands:
    def __init__(self):
        # Statements for Data Package records. Accessed from HTTP server
        self.CREATEDPTABLE = ("CREATE TABLE IF NOT EXISTS DataPackages"
        "(PrimaryKey          INTEGER  PRIMARY KEY ON CONFLICT FAIL AUTOINCREMENT UNIQUE ON CONFLICT FAIL,"
        "UID                STRING,"
        "Name               STRING,"
        "Hash               VARCHAR(300),"
        "SubmissionDateTime DATETIME DEFAULT (CURRENT_TIMESTAMP),"
        "SubmissionUser     STRING,"
        "CreatorUid         STRING,"
        "Keywords           CHAR     DEFAULT foobar,"
        "MIMEType           STRING   DEFAULT [application/x-zip-compressed],"
        "Size               INTEGER,"
        "Privacy            INTEGER  DEFAULT 0);")
        self.MISSIONUPLOADCALLSIGN = "SELECT Callsign FROM Users WHERE UID=?"
        self.INSERTDPINFO = "INSERT INTO DataPackages (UID, Name, Hash, SubmissionUser, CreatorUid, Size) VALUES (?,?,?,?,?,?);"
        self.ROWBYHASH = "SELECT * FROM DataPackages WHERE Hash=?"
        self.SELECTALLDP = "SELECT * FROM DataPackages WHERE Privacy = 0"
        self.RETRIEVECALLSIGNFROMUID = "SELECT Callsign FROM Users WHERE Uid = ?"

        # Statements to work with storing/retrieving video links. Accessed from HTTP server
        self.CREATEVIDEOTABLE = ("CREATE TABLE IF NOT EXISTS VideoLinks"
        "(PrimaryKey         INTEGER PRIMARY KEY ON CONFLICT FAIL AUTOINCREMENT UNIQUE ON CONFLICT FAIL,"
        "FullXmlString       STRING,"
        "Timestamp           DATETIME DEFAULT (CURRENT_TIMESTAMP),"
        "Protocol            STRING,"
        "Alias               STRING,"
        "Uid                 STRING,"
        "Address             STRING,"
        "Port                INTEGER DEFAULT -1,"
        "RoverPort           INTEGER DEFAULT -1,"
        "IgnoreEmbeddedKlv   STRING DEFAULT false,"
        "PreferredMacAddress STRING DEFAULT NULL,"
        "Path                STRING DEFAULT NULL,"
        "Buffer              INTEGER DEFAULT -1,"
        "Timeout             INTEGER,"
        "RtspReliable        INTEGER DEFAULT 0);")
        self.INSERTVIDEO = "INSERT INTO VideoLinks (FullXmlString,Protocol,Alias,Uid,Address,Port,RoverPort,IgnoreEmbeddedKlv,PreferredMacAddress,Path,Buffer,Timeout,RtspReliable) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"
        self.GETALLVIDEOS = "SELECT FullXmlString FROM VideoLinks"
        self.GETVIDEOSWITHUID = "SELECT FullXmlString from VideoLinks WHERE Uid=?"

        # Statements for User records accessed from HTTP server
        self.CREATEUSERTABLE = ("CREATE TABLE IF NOT EXISTS Users"
        "(PrimaryKey         INTEGER PRIMARY KEY ON CONFLICT FAIL AUTOINCREMENT UNIQUE ON CONFLICT FAIL,"
        "Uid                 STRING,"
        "Callsign             STRING)")
        self.ADDUSER = ("INSERT INTO Users (Uid, Callsign) VALUES (?, ?)")
        self.REMOVEUSER = ("DELETE FROM Users WHERE Uid = ?")


#######################################################
# 
# ClientInformationController.py
# Python implementation of the Class ClientInformationController
# Generated by Enterprise Architect
# Created on:      19-May-2020 6:56:00 PM
#
#######################################################
from lxml import etree
from FreeTAKServer.controllers.BasicModelInstantiate import BasicModelInstantiate
import uuid
from FreeTAKServer.controllers.configuration.LoggingConstants import LoggingConstants
from FreeTAKServer.controllers.CreateLoggerController import CreateLoggerController
from FreeTAKServer.model.FTSModel.Event import Event
from FreeTAKServer.model.ClientInformation import ClientInformation
from FreeTAKServer.controllers.XMLCoTController import XMLCoTController

logger = CreateLoggerController("ClientInformationController").getLogger()

loggingConstants = LoggingConstants()

class ClientInformationController(BasicModelInstantiate):
    def __init__(self):
        pass
    '''
    connection setup is obsolete with intstantiateClientInformationModelFromController
    '''

    def intstantiateClientInformationModelFromConnection(self, rawClientInformation, queue):
        try:
            tempObject = Event.Connection()
            self.clientInformation = ClientInformation()
            argument = "initialConnection"
            self.clientInformation.dataQueue = queue
            self.clientInformation.socket = rawClientInformation.socket
            self.clientInformation.IP = rawClientInformation.ip
            self.clientInformation.idData = rawClientInformation.xmlString
            self.clientInformation.alive = 1
            self.clientInformation.ID = uuid.uuid1().int
            self.clientInformation.modelObject = XMLCoTController().serialize_CoT_to_model(tempObject, etree.fromstring(rawClientInformation.xmlString.encode()))
            return self.clientInformation
        except Exception as e:
            logger.debug('error in client information controller '+str(e))
            return -1


from sqlalchemy import Column
from FreeTAKServer.model.SQLAlchemy.Root import Base, Root
from sqlalchemy import String
from sqlalchemy.orm import relationship
from FreeTAKServer.model.SQLAlchemy.ExCheck import ExCheck
from sqlalchemy import DateTime
from sqlalchemy import Integer
from sqlalchemy import ForeignKey

class ExCheckChecklist(Base):
    __tablename__ = "ExCheckChecklist"
    PrimaryKey = Column(Integer, primary_key = True, autoincrement=True)
    startTime = Column(DateTime)
    creatorUid = Column(String)
    description = Column(String)
    name = Column(String)
    callsign = Column(String)
    uid = Column(String)
    filename = Column(String)
    template_id = Column(Integer, ForeignKey('ExCheck.PrimaryKey'))


from .SendCoTAbstractController import SendCoTAbstractController
from FreeTAKServer.model.SpecificCoT.SendTakPong import SendTakPong


class SendTakPongController(SendCoTAbstractController):
    def __init__(self, RawCoT):
        tempObject = super().Event.takPong()
        object = SendTakPong()
        self.fill_object(object, tempObject, RawCoT)


import multiprocessing

class AddDataToCoTList:
    def __init__(self):
        pass

    #this function sends specified data to all pipes within a provided array
    def send(self, pipes, data):
        for pipe in pipes:
            try:
                pipe.send(data)
            except Exception as e:
                print(e)
                pass
        return 1

    #this function attempts to receive data from a specified pipe and then return the data
    def recv(self, pipe):
        try:
            out = pipe.poll(timeout=0.1)
            if out:
                data = pipe.recv()
                return data
            else:
                return 0
        except Exception as e:
            print(e)


from FreeTAKServer.controllers.configuration.MainConfig import MainConfig

class DatabaseConfiguration:
    DataBasePath = MainConfig.DBFilePath
    DataBaseType = str('sqlite:///')
    DataBaseConnectionString = str(DataBaseType+DataBasePath)
    #DataBaseConnectionString = str('sqlite:////home/ghost/FTSTesting/FreeTAKServer/controllers/testing.db')



from .SSLCoTServiceVariables import SSLCoTServiceVariables as vars

class SSLCoTService:
    def __init__(self, CoTServiceIP = vars().SSLCoTServiceIP,
                 CoTServicePort = vars().SSLCoTServicePort,
                 CoTServiceStatus = vars().SSLCoTServiceStatus):
        self.SSLCoTServiceIP = CoTServiceIP
        self.SSLCoTServicePort = CoTServicePort
        self.SSLCoTServiceStatus = CoTServiceStatus


from .SpecificCoTAbstract import SpecificCoTAbstract

class SendOther(SpecificCoTAbstract):
    def __init__(self):
        self.define_variables()
        self.setType("other")
        self.martiPresent = False

    def setMartiPresent(self, martiPresent):
        self.martiPresent = martiPresent

    def getMartiPresent(self):
        return self.martiPresent


#######################################################
# 
# Archive.py
# Python implementation of the Class Archive
# Generated by Enterprise Architect
# Created on:      26-Sep-2020 9:41:27 PM
# Original author: natha
# 
#######################################################
from sqlalchemy import Column, ForeignKey
from FreeTAKServer.model.SQLAlchemy.Root import Base
from sqlalchemy.orm import relationship


class Archive(Base):

    __tablename__ = 'Archive'
    PrimaryKey = Column(ForeignKey("Detail.PrimaryKey"), primary_key=True)
    Detail = relationship("Detail", back_populates="archive")


#######################################################
# 
# ClientSendHandler.py
# Python implementation of the Class ClientSendHandler
# Generated by Enterprise Architect
# Created on:      21-May-2020 9:46:48 AM
# Original author: Natha Paquette
# 
#######################################################


class ClientSendHandler:
# default constructor  def __init__(self):  

    def attemptToSendData():
        pass

    def monitorDataQueue():
        pass

    def serializeModelObject():
        pass


from FreeTAKServer.model.SQLAlchemy.ExCheck import ExCheck
from FreeTAKServer.model.SQLAlchemy.ExCheck import ExCheckData
from FreeTAKServer.model.ExCheck.templateInstanceContents import templateInstanceContents
from FreeTAKServer.controllers.DatabaseControllers.table_controllers import TableController
from FreeTAKServer.model.SQLAlchemy.ExCheckKeywords import ExCheckKeywords
from sqlalchemy.sql import text
unsuportedAttribs = ['_sa_instance_state']
#TODO: move out all serializers into a serializers controller

class ExCheckController(TableController):

    def __init__(self):
        self.table = ExCheck

    def queryChildren(self, query, columns, session):
        from FreeTAKServer.model.SQLAlchemy.ExCheckData import ExCheckData
        output = session.query(self.table).join(ExCheckData, self.table.data).filter(text(query)).all()
        return output

    def create(self, session, object):
        row = self.convert_exCheck_model_to_row(object)
        session.add(row)
        session.commit()

    def convert_exCheck_model_to_row(self, modelObject):
        if isinstance(modelObject, templateInstanceContents):
            exCheck_row = ExCheck()
            for attribName, attribValue in modelObject.__dict__.items():
                if attribName not in unsuportedAttribs:
                    if attribName == 'data':
                        data_row = ExCheckData()
                        for attribName, attribValue in attribValue.__dict__.items():
                            if attribName not in unsuportedAttribs:
                                if isinstance(attribValue, list):
                                    keyword_row = ExCheckKeywords()
                                    keyword_row.name = attribValue[0]
                                    keyword_row.description = attribValue[1]
                                    keyword_row.callsign = attribValue[2]
                                    setattr(data_row, attribName, keyword_row)

                                else:
                                    setattr(data_row, attribName, attribValue)

                        setattr(exCheck_row, 'data', data_row)
                    else:
                        setattr(exCheck_row, attribName, attribValue)
                else:
                    pass
            return exCheck_row
        else:
            raise TypeError('incorrect object passed to convert_exCheck_model_to_row')


from FreeTAKServer.controllers.services.FederationServiceController import FederationServiceController
from FreeTAKServer.controllers.services.FederationServiceAbstract import FederationServiceAbstract
import ssl
import multiprocessing
import socket
import threading
from FreeTAKServer.controllers.configuration.LoggingConstants import LoggingConstants
from FreeTAKServer.controllers.CreateLoggerController import CreateLoggerController
loggingConstants = LoggingConstants()
logger = CreateLoggerController("FederationServer").getLogger()

class FederationServerServiceController(FederationServiceAbstract):

    """this service acts as both an orchestrator for the connection of new clients and
    an abstraction layer between FTS and all clients for the transfer of data. and is
    always running.
    """

    def listen_for_connection(self, ip, port, ssock):
        '''
        listens for connection
        :param ip:
        :param port:
        '''
        while True:
            try:
                conn, addr = ssock.accept()
                federate = self.connected(ip, port, conn)
            except Exception as e:
                logger.error("there has been an exception in listening for connections "
                             "within the reception of Federates " + str(e))

    def connected(self, ip, port, conn):
        try:
            '''
            this function handles connecting federates
            :return:
            '''
            federate = FederationServiceController(ip, port, conn).start()
            self.federateClients.add_client(federate)
            return federate
        except Exception as e:
            logger.error('there has been an exception in the '
                         'processing of a client connection'
                         ' within the connect function ' + str(e))

    def start(self, ip, port):
        try:
            '''
            the main start function which controls all other threads
            :param ip:
            :param port:
            :return:
            '''
            ssock = self._create_socket(ip, port)
            self.data_handler_thread = threading.Thread(target=self.data_handler, args=())
            self.data_handler_thread.start()
            self.listen_for_connection(ip, port, ssock)
        except Exception as e:
            logger.error('there has been an exception in the starting of the server'
                         ' Federation Service ' + str(e))

    def _create_socket(self, ip, port):
        try:
            # TODO: move certificate location and password to configuration
            context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            context.load_cert_chain('/home/natha/ProgrammerStuff/FTS/FTSDev/FreeTAKServer/Certs/ServerCerts/server.crt',
                                    '/home/natha/ProgrammerStuff/FTS/FTSDev/FreeTAKServer/Certs/ServerCerts/server.key',
                                    password='atakatak')
            # context.load_cert_chain(r'C:\Users\natha\Documents\programmer_stuff\playground\CERT\TAKServerCerts\FTS certs\fts-server-cert.pem', r'C:\Users\natha\Documents\programmer_stuff\playground\CERT\TAKServerCerts\FTS certs\fts-server-key.key',password='ftsfts')
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
            sock.bind((ip, port))
            sock.listen(5)
            ssock = context.wrap_socket(sock, server_side=True)
            return ssock
        except Exception as e:
            logger.error('there has been an exception in the creation'
                         ' of the server socket ' + str(e))

    def data_handler(self):
        '''
        this function controlls the sending and receiving of data
        via the pipe passed in startup and is run in a thread
        :return:
        '''
        while True:
            try:
                self.pipe.poll(timeout = 0.01)
                success = self.send_data_to_clients(self.pipe.get(), self.federateClients.get_all_clients())
            except multiprocessing.TimeoutError:
                pass
            except Exception as e:
                logger.error('there has been an exception in the sending of'
                             ' data from pipe to federate clients ' + str(e))

            try:
                success = self.recv_data_from_clients(self.federateClients.get_all_clients())
            except Exception as e:
                logger.error('there has been an exception in the reception of'
                             ' data from pipe to federate clients ' + str(e))



#######################################################
# 
# takv.py
# Python implementation of the Class takv
# Generated by Enterprise Architect
# Created on:      28-Sep-2020 10:48:20 PM
# Original author: natha
# 
#######################################################
from sqlalchemy import Column, ForeignKey
from FreeTAKServer.model.SQLAlchemy.Root import Base
from sqlalchemy import String
from sqlalchemy.orm import relationship


class Takv(Base):
# default constructor  def __init__(self):
    __tablename__ = "Takv"
    PrimaryKey = Column(ForeignKey("Detail.PrimaryKey"), primary_key=True)
    Detail = relationship("Detail", back_populates="takv")
    # type of physical device
    device = Column(String)
    # the operating system running TAK
    os = Column(String)
    # the variant of TAK
    platform = Column(String)
    # the version of TAK running on the device
    version = Column(String)


from sqlalchemy import Column
from FreeTAKServer.model.SQLAlchemy.Root import Base, Root
from sqlalchemy import String
from sqlalchemy.orm import relationship
from FreeTAKServer.model.SQLAlchemy.ExCheckData import ExCheckData
from sqlalchemy import DateTime
from sqlalchemy import Integer
from sqlalchemy import ForeignKey


class ExCheck(Base):
    __tablename__ = 'ExCheck'
    PrimaryKey = Column(Integer, primary_key = True, autoincrement=True)
    timestamp = Column(DateTime)
    creatorUid = Column(String)
    checklist = relationship("ExCheckChecklist", backref='template')
    data = relationship("ExCheckData", uselist=False, cascade="all, delete", backref="ExCheck")



import os
from FreeTAKServer.controllers.configuration.DataPackageServerConstants import DataPackageServerConstants
from FreeTAKServer.controllers.configuration.LoggingConstants import LoggingConstants
from pathlib import PurePath
class CreateStartupFilesController:
    def __init__(self):
        self.file_dir = os.path.dirname(os.path.realpath(__file__))
        self.dp_directory = PurePath(self.file_dir, DataPackageServerConstants().DATAPACKAGEFOLDER)
        self.logs_directory = PurePath(self.file_dir, LoggingConstants().LOGDIRECTORY)
        self.createFolder()
    def createFolder(self):
        try:
            os.mkdir(self.dp_directory)
        except:
            pass
        try:
            os.mkdir(self.logs_directory)
        except:
            pass
        ERRORLOG = open(LoggingConstants().ERRORLOG, "w")
        ERRORLOG.close()
        HTTPLOG = open(LoggingConstants().HTTPLOG, "w")
        HTTPLOG.close()
        DEBUGLOG = open(LoggingConstants().DEBUGLOG, "w")
        DEBUGLOG.close()
        INFOLOG = open(LoggingConstants().INFOLOG, "w")
        INFOLOG.close()



from FreeTAKServer.model.FTSModel.fts_protocol_object import FTSProtocolObject
#######################################################
# 
# chatgrp.py
# Python implementation of the Class chatgrp
# Generated by Enterprise Architect
# Created on(FTSProtocolObject):      11-Feb-2020 11(FTSProtocolObject):08(FTSProtocolObject):10 AM
# Original author: Corvo
# 
#######################################################
from FreeTAKServer.model.FTSModelVariables.ChatgrpVariables import ChatgrpVariables as vars

class Chatgrp(FTSProtocolObject):
    def __init__(self):
        self.uid0 = None
        self.uid1 = None
        self.id = None

    @staticmethod
    def geochat(UID0 = vars.geochat().UID0, UID1 = vars.geochat().UID1, ID = vars.geochat().ID):
        chatgrp = Chatgrp()
        chatgrp.setid(ID)
        chatgrp.setuid0(UID0)
        chatgrp.setuid1(UID1)
        return chatgrp

    def chatToTeamFunc(self, uid0, uid1, id):
        self.setuid0(uid0)
        self.setuid1(uid1)
        self.setid(id)

    def chatToGroupFunc(self, uid0, uid1, id):
        self.setuid0(uid0)
        self.setuid1(uid1)
        self.setid(id)

    def chatToAllFunc(self, uid0, uid1, id):
        self.setuid0(uid0)
        self.setuid1(uid1)
        self.setid(id)

    def getuid0(self):
        return self.uid0

    # uid0 setter
    def setuid0(self, uid0=0):
        self.uid0=uid0

    # uid1 getter
    def getuid1(self):
        return self.uid1

    # uid1 setter
    def setuid1(self, uid1=0):
        self.uid1=uid1

    # id getter
    def getid(self):
        return self.id

    # id setter
    def setid(self, id=0):
        self.id=id
     
     



from FreeTAKServer.model.FTSModel.fts_protocol_object import FTSProtocolObject
#######################################################
# 
# remarks.py
# Python implementation of the Class remarks
# Generated by Enterprise Architect
# Created on(FTSProtocolObject):      11-Feb-2020 11(FTSProtocolObject):08(FTSProtocolObject):08 AM
# Original author: Corvo
# 
#######################################################
from FreeTAKServer.model.FTSModelVariables.RemarksVariables import RemarksVariables as vars
import datetime as dt

class Remarks(FTSProtocolObject):
  """Provides a place to annotate CoT with free text information.  e.g. comments
  from other users about the current COT. Used also fro the geoChat.
  the xml body of this class is used to transport the chat message
  """
  def __init__(self):
    self.time = None
    self.source = None
    self.sourceID = None
    self.to = None
    self.INTAG = None

  @staticmethod
  def connection():
    remarks = Remarks()
    return remarks

  @staticmethod
  def FederatedCoT():
    remarks = Remarks()
    return remarks

  @staticmethod
  def drop_point(TIME = vars.drop_point().TIME, SOURCE = vars.drop_point().SOURCE, SOURCEID = vars.geochat().SOURCEID, INTAG = vars.geochat().INTAG):
    remarks = Remarks()
    remarks.settime(time=TIME)
    remarks.setsource(source=SOURCE)
    remarks.setsourceID(SOURCEID)
    remarks.setINTAG(INTAG)
    return remarks

  @staticmethod
  def geochat(TIME = vars.geochat().TIME, SOURCE = vars.geochat().SOURCE, SOURCEID = vars.geochat().SOURCEID, INTAG = vars.geochat().INTAG, TO = vars.geochat().TO):
    remarks = Remarks()
    remarks.settime(time=TIME)
    remarks.setsource(source=SOURCE)
    remarks.setsourceID(SOURCEID)
    remarks.setINTAG(INTAG)
    remarks.setto(TO)
    return remarks

  # time getter
  def gettime(self):
    return self.time

  # time setter
  def settime(self, time=0):
    DATETIME_FMT = "%Y-%m-%dT%H:%M:%SZ"
    if time == None:
      timer = dt.datetime
      now = timer.utcnow()
      zulu = now.strftime(DATETIME_FMT)
      self.time = zulu
    else:
      self.time = time

  # to getter
  def getto(self):
    return self.to

  # to setter
  def setto(self, to=0):
    self.to=to

  # source getter
  def getsource(self):
    return self.source

  # source setter
  def setsource(self, source=0):
    self.source=source

  def getsourceID(self):
    return self.sourceID

  # source setter
  def setsourceID(self, sourceID=0):
    self.sourceID = sourceID
    
  def setINTAG(self, INTAG):
    self.INTAG = INTAG
    
  def getINTAG(self):
    return self.INTAG


#######################################################
# 
# Summary.py
# Python implementation of the Class Summary
# Generated by Enterprise Architect
# Created on:      26-Sep-2020 9:41:46 PM
# Original author: natha
# 
#######################################################
from sqlalchemy import Column
from FreeTAKServer.model.SQLAlchemy.Root import Base
from sqlalchemy import String
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship

class Summary(Base):
# default constructor  def __init__(self):
    __tablename__ = "Summary"

    PrimaryKey = Column(ForeignKey("Detail.PrimaryKey"), primary_key=True)
    Detail = relationship("Detail", back_populates="summary")
    INTAG = Column(String)


#######################################################
# 
# Chat.py
# Python implementation of the Class Chat
# Generated by Enterprise Architect
# Created on:      04-Nov-2020 4:26:46 PM
# Original author: natha
# 
#######################################################


class Chat:
# default constructor  def __init__(self):  
    # the text of the GeoChat message
    message = ""
    # the name of the chat's sender, changing this will also change the chat room for
    # the client.
    sender = ""


from FreeTAKServer.controllers.serializers.serializer_abstract import SerializerAbstract
from FreeTAKServer.model.FTSModel.Event import Event
from typing import NewType, List
from lxml import etree
from FreeTAKServer.model.FTSModel.fts_protocol_object import FTSProtocolObject
import time
class XmlSerializer(SerializerAbstract):

    __exception_mapping_dict = {'_group': '__group', '_serverdestination': '__serverdestination', '__group': "_group", '__serverdestination': "_serverdestination", "chat": "__chat", "__chat": "chat"}

    __ftsObjectType = NewType('ftsObject', FTSProtocolObject)

    def from_format_to_fts_object(self, object: str, FTSObject: Event) -> Event:
        for element in etree.XML(object).iter():
            if element.text is not None:
                setters = self._get_fts_object_var_setter(FTSObject, "INTAG")
                setter = self._get_method_in_method_list(setters, element.tag)
                setter(element.text)
            for key, var in element.attrib.items():
                setters = self._get_fts_object_var_setter(FTSObject, key)
                setter = self._get_method_in_method_list(setters, element.tag)
                setter(var)
        return FTSObject

    def from_fts_object_to_format(self, FTSObject: Event) -> etree._Element:
        object_body = self._from_fts_object_to_format_body(FTSObject)
        root = object_body
        for element in root.iter():
            if element.text == 'DATA':
                getters = self._get_fts_object_var_getter(FTSObject, "INTAG")
                getter = self._get_method_in_method_list(getters, element.tag)
                tempvar = getter()
                element.text = tempvar
            for key, var in element.attrib.items():
                getters = self._get_fts_object_var_getter(FTSObject, key)
                getter = self._get_method_in_method_list(getters, element.tag)
                tempvar = getter()
                if tempvar != None:
                    element.attrib[key] = str(tempvar)
        y = etree.tostring(root)
        return root

    def _get_method_in_method_list(self, method_list: List[callable], expected_class_name: str) -> callable:
        if expected_class_name.lower() in self.__exception_mapping_dict:
            expected_class_name = self.__exception_mapping_dict[expected_class_name.lower()]
        if len(method_list) > 0:
            for method in method_list:
                # required due to pythons privacy conventions
                if method.__self__.__class__.__name__.lower() == expected_class_name.lower():
                    return method
                else:
                    pass
            raise AttributeError(expected_class_name + ' does not have specified attribute')
        else:
            raise AttributeError(expected_class_name + ' does not have specified attribute')

    def _from_fts_object_to_format_body(self, FTSObject: __ftsObjectType) -> etree._Element:
        """ converts FTS object to xml outline without data

        this method takes an FTSObject instance and converts it to an xml message
        without populating it with any actual data.

        :param FTSObject an instance of FTSObject to be serialized

        :return an etree element containing the outline of the xml object
        """
        # set object name to lowercase
        name = type(FTSObject).__name__
        name = name[0].lower() + name[1:]
        if name.lower() in self.__exception_mapping_dict:
            name = self.__exception_mapping_dict[name.lower()]
        # create object and get attributes as dict
        xmlobj = etree.Element(name)
        variables = vars(FTSObject)
        # iterate variables and create attributes
        for key, value in variables.items():
            if '_'+FTSObject.__class__.__name__.lower()+'__' in key.lower():
                continue
            if issubclass(type(value), FTSProtocolObject):
                try:
                    xmlobj.append(self._from_fts_object_to_format_body(value))
                except AttributeError:
                    pass
            # adds dummy data to indicate that intag data needs to be added for from_fts_object_to_format
            elif key == "INTAG":
                xmlobj.text = "DATA"
            # continues iteration over list type attributes
            elif isinstance(value, list):
                for obj in value:
                    xmlobj.append(self._from_fts_object_to_format_body(obj))
            # handles all other regular attributes
            elif value is not None:
                xmlobj.attrib[key] = ""
        z = etree.tostring(xmlobj)
        return xmlobj


'''xmlstring = '<event version="2.0" uid="ANDROID-R5CN70EYKQH" type="a-f-G-U-C" how="h-e" start="2020-12-24T18:16:22.325Z" time="2020-12-24T18:16:22.325Z" stale="2020-12-24T18:22:37.325Z"><detail><__group name="Teal" role="Team Member"/><status battery="76"/><takv version="4.2.0.4 (47e136dd).1607456856-CIV" platform="ATAK-CIV" device="SAMSUNG SM-N986U" os="29"/><track course="159.1462509079387" speed="0.0"/><contact callsign="SPAC3SLOTH" endpoint="*:-1:stcp" /><uid Droid="SPAC3SLOTH"/><precisionlocation altsrc="GPS" geopointsrc="GPS"/></detail><point le="9999999.0" ce="11.0" hae="178.84407323983876" lon="-76.675505" lat="39.664392"/></event>'
fts_obj = xml_serializer().from_format_to_fts_object(xmlstring, Event.Connection())
obj = xml_serializer().from_fts_object_to_format_body(Event.Connection())
obj = xml_serializer().from_fts_object_to_format(fts_obj, etree.tostring(obj))
print(etree.tostring(obj))
print(etree.tostring(obj).decode() == xmlstring)
finish = time.time()
print(finish-start)'''


from lxml import etree
from FreeTAKServer.controllers.configuration.LoggingConstants import LoggingConstants
from FreeTAKServer.controllers.CreateLoggerController import CreateLoggerController
from FreeTAKServer.model.FTSModel.Event import Event as event
from FreeTAKServer.controllers.configuration.MainConfig import MainConfig
from abc import ABC
from FreeTAKServer.controllers.serializers import xml_serializer

loggingConstants = LoggingConstants()
logger = CreateLoggerController("SendCoTAbstract").getLogger()

class SendCoTAbstractController(ABC):
    Event = event
    def __init__(self, object):
        pass

    def fill_object(self, object, tempObject, RawCoT, addToDB = MainConfig.SaveCoTToDB):
        try:
            object.modelObject = self.create_model_object(tempObject, RawCoT.xmlString)

        except Exception as e:
            if tempObject.type != "other":
                self.handel_serialization_exception(object, RawCoT)
                logger.error('an undocumented CoT has been sent returning the error ' + str(e) + ' please post this as an issue on the FTS git so we can document and fix it')
                return 1
            else:
                logger.warning('there has been an exception in the creation of the model object ' + str(e))
                return -1

        try:
            object.xmlString = self.create_xml_string(object.modelObject)
        except Exception as e:
            logger.warning('there has been an exception in the creation of a xmlString ' + str(e))
        try:
            if addToDB == True:
                RawCoT.dbController.create_CoT(object.modelObject)
            else:
                pass
        except Exception as e:
            logger.warning('there has been an exception in the creation of a database instance of this object ' + str(e))

        self.setObject(object)

    def create_model_object(self, tempObject, xmlString):
        """
        this function takes an empty model object and xml as arguments and then calls the serializer within the xmlcotcontroller
        module which returns a model object occupied with all the data from the CoT
        """
        from FreeTAKServer.controllers.XMLCoTController import XMLCoTController

        modelInstance = xml_serializer.XmlSerializer().from_format_to_fts_object(xmlString, tempObject)
        return modelInstance
    def create_xml_string(self, modelObject):
        """
        this function calls the model to xml serializer within XMLCoTController
        with an instantiated model object supplied in the function argument
        """
        from FreeTAKServer.controllers.XMLCoTController import XMLCoTController
        try:
            xml = XMLCoTController().serialize_model_to_CoT(modelObject, 'event')
            return xml
        except Exception as e:
            logger.error('there has been an exception in the creation of the xml string ' +str(e))
            return -1

    def handel_serialization_exception(self, object, RawCoT):
        """
        this function will handle any exception thrown in the process of serialization of an xml CoT to it's respective model object
        by processing it as an Other type
        """
        from FreeTAKServer.controllers.SpecificCoTControllers.SendOtherController import SendOtherController
        object = SendOtherController(RawCoT).getObject()
        object.clientInformation = RawCoT.clientInformation
        self.setObject(object)


    def setObject(self, object):
        self.Object = object

    def getObject(self):
        return self.Object

    def reloadXmlString(self):
        self.Object.xmlString = self.create_xml_string(self.Object.modelObject)


from FreeTAKServer.model.FTSModel.fts_protocol_object import FTSProtocolObject
#######################################################
# 
# Contact.py
# Python implementation of the Class Contact
# Generated by Enterprise Architect
# Created on(FTSProtocolObject):      11-Feb-2020 11(FTSProtocolObject):08(FTSProtocolObject):08 AM
# Original author: Corvo
# 
#######################################################
from FreeTAKServer.model.FTSModelVariables.ContactVariables import ContactVariables as vars

class Contact(FTSProtocolObject):
    def __init__(self):
        self.callsign = None 
        self.endpoint = None
        self.iconsetpath = None
        self.uid = None
        self.name = None

    @staticmethod
    def drop_point(CALLSIGN = vars.drop_point().CALLSIGN):
        contact = Contact()
        contact.setcallsign(callsign=CALLSIGN)
        return contact

    @staticmethod
    def connection(CALLSIGN = vars.connection().CALLSIGN, ENDPOINT = vars.connection().ENDPOINT,
                   ICONSETPATH = vars.connection().ICONSETPATH,UID = vars.connection().UID,
                   NAME = vars.connection().NAME):

        contact = Contact()
        contact.setcallsign(CALLSIGN)
        contact.setname(NAME)
        contact.setuid(UID)
        contact.setendpoint(ENDPOINT)
        contact.seticonsetpath(ICONSETPATH)
        return contact

    @staticmethod
    def geochat(CALLSIGN=vars.geochat().CALLSIGN, ENDPOINT=vars.geochat().ENDPOINT,
                   ICONSETPATH=vars.geochat().ICONSETPATH, UID=vars.geochat().UID,
                   NAME=vars.geochat().NAME):
        contact = Contact()
        contact.setcallsign(CALLSIGN)
        contact.setname(NAME)
        contact.setuid(UID)
        contact.setendpoint(ENDPOINT)
        contact.seticonsetpath(ICONSETPATH)
        return contact

    @staticmethod
    def emergency_on(CALLSIGN=vars.emergency_on().CALLSIGN, ENDPOINT=vars.emergency_on().ENDPOINT,
                ICONSETPATH=vars.emergency_on().ICONSETPATH, UID=vars.emergency_on().UID,
                NAME=vars.emergency_on().NAME):
        contact = Contact()
        contact.setcallsign(CALLSIGN)
        contact.setname(NAME)
        contact.setuid(UID)
        contact.setendpoint(ENDPOINT)
        contact.seticonsetpath(ICONSETPATH)
        return contact

    @staticmethod
    def UserUpdate(ENDPOINT=vars.UserUpdate().endpoint, PHONE=vars.UserUpdate().phone,
                   CALLSIGN=vars.UserUpdate().callsign):
        contact = Contact()
        contact.setendpoint(ENDPOINT)
        contact.setphone(PHONE)
        contact.setcallsign(CALLSIGN)
        return contact
    
    @staticmethod
    def SimpleCoT(CALLSIGN = vars.SimpleCoT().callsign):
        contact = Contact()
        contact.setcallsign(CALLSIGN)
        return contact

    @staticmethod
    def Presence(CALLSIGN = vars.Presence().callsign):
        contact = Contact()
        contact.setcallsign(CALLSIGN)
        return contact

    @staticmethod
    def SimpleCoT(CALLSIGN = vars.SimpleCoT().callsign):
        contact = Contact()
        contact.setcallsign(CALLSIGN)
        return contact

    @staticmethod
    def Presence(CALLSIGN = vars.Presence().callsign):
        contact = Contact()
        contact.setcallsign(CALLSIGN)
        return contact

    # iconsetpath getter 
    def geticonsetpath(self): 
        return self.iconsetpath 
 
    # iconsetpath setter 
    def seticonsetpath(self, iconsetpath=None):
        self.iconsetpath=iconsetpath 
 
    # callsign getter 
    def getcallsign(self): 
        return self.callsign 
 
    # callsign setter 
    def setcallsign(self, callsign=None):
        self.callsign=callsign 
 
     
    # endpoint getter 
    def getendpoint(self): 
        return self.endpoint 
 
    # endpoint setter 
    def setendpoint(self, endpoint=None):
        self.endpoint=endpoint

    def getuid(self):
        return self.uid

        # uid setter 

    def setuid(self, uid=None):
        self.uid = uid 
        
    def getname(self):
        return self.name

        # name setter 

    def setname(self, name=None):
        self.name = name

    def getphone(self):
        return self.phone

    def setphone(self, phone=None):
        self.phone = phone



#######################################################
# 
# RootController.py
# Python implementation of the Class TableController
# Generated by Enterprise Architect
# Created on:      24-Sep-2020 4:49:08 PM
# Original author: natha
# 
#######################################################


class RootController:
	def __init__(self):  
		self.table = None
	
	def delete(self, session, query):
		# this function removes a row from the datapackages table based on the uid
		session.query(self.table).filter(text(query)).delete()
		session.commit()

	def create(self, session, **args):
		# this function adds a new row to the datapackages table
		session.add(self.table(**args))
		session.commit()

	def query(self, session, query, columns):
		# query needs to be applicable to datapackage object tuple(['DataPackage.'+x for x in columns])
		output = session.query(*tuple([getattr(self.table, x) if x != '*' else DataPackage for x in columns])).filter(text(query)).all()
		return output

	def update(self, session, query, column_value):
		DataPackages = session.query(self.table).filter(text(query)).all() #self.query(session, query, [column for column, value in column_value.items()])
		for dp in DataPackages:
			for column, value in column_value.items():
				setattr(dp, column, value)
		session.commit()